#!/usr/bin/perl
#
# This script is intended to add the signatures based on the rules
# together with the references to the database. This should speed
# up any output plugin (or something similar) using the database. 
#
# Author: Dirk Geschke <Dirk@geschke-online.de>
#

use DBI;
use Getopt::Std;

my $dbh;

getopts('c:C:P:R:h');

if ($opt_h)
{
 print "\nUsage:";
 print "\n        $0 [-h] [-c <config.file>] [-C <classification.config>] ";
 print "\n        [-P priorityrange] [-R <reference.config>] file1.rule file2.rule ...\n";
 print "\n -c <config.file>           : name of the configuration file for $0.";
 print "\n -C <classification.config> : name of the classification file";
 print "\n -h                         : this list";
 print "\n -P priorityrange           : Range of priorities associated with each rule";
 print "\n -R <reference.config>      : name of the reference file"; 
 print "\n file1.rule file2.rule ...  : list of rule files to be used";
 print "\n\n";
 exit 0;
}

$conffile=$opt_c || "rules.pl.conf";

print "Configuration file: $conffile\n";

do "$conffile";

if ($opt_C)
{
 $ClassFile=$opt_C;
}
#
# check if classfile exists
#
open(CLASSFILE,"< $ClassFile") or die "Can't open CLASSFILE <$ClassFile>: $!";

if ($opt_R)
{
 $ReferenceFile=$opt_R;
}
#
# check if reference file exists
#
open(REFFILE,"< $ReferenceFile") or die "Can't open REFFILE <$ReferenceFile>: $!";

#
# set the priority range
#
if ($opt_P)
{
  $Priority=0+$opt_P;
}

if ($dbtype eq "mysql")
{
  $dbh = DBI->connect("DBI:mysql:dbname=".$dbname, $dbuser , $dbpass ,
			 { RaiseError => 1, AutoCommit => 1 });
}
elsif ($dbtype eq "postgres")
{
  $dbh = DBI->connect("DBI:Pg:dbname=".$dbname, $dbuser , $dbpass ,
			 { RaiseError => 1, AutoCommit => 1 });
}
else
{
  print "Databastype <$dbtype> is not supported.\n";
  exit 1;
}
#my $dbh = DBI->connect("DBI:Pg:dbname=snort", "snort" , "" ,{ RaiseError => 1, AutoCommit => 1 });


#
# if set $PrioRange is greater zero then all signatures are inserted in
# the database from 1 to $PrioRange. Normally only the default priority
# is inserted, so if it later is changed it will be missed in the database
# (if you use DBTrust, otherwise it will be shown with the old priority).
# NOTE: Normally the priority is not set directly via the rule. Instead
#       the priority is taken from the classification.config file.
# -> Config file: my $PrioRange=0;
my $sql;
my $sth;

if ($PrioRange>0)
{
    print "Setting priority to range 1..$PrioRange\n";
}

my $classnr=1;
while (<CLASSFILE>) 
  {
    if (m/^config/) 
      {
         chop();
         $class=(split(/:/,$_))[1] ;
         ($class,$classdesc[$classnr],$prio[$classnr]) = 
               (split(/,/,$class))[0,1,2];
	 while ( substr($class,0,1) eq ' ' ) 
	 {
	     $class=substr($class,1,length($class)-1);
	 }
	 $class =~ tr/A-Z/a-z/;
         $classshort[$classnr]=$class;
#	 print "<$classshort[$classnr]>\n";
	 $classnr++;
       }
  }
close(CLASSFILE);

my $refnr=0;
while (<REFFILE>) {
  if (m/^config/) {
     chop();
     ($reftype[$refnr],$url[$refnr])=(split(/\s+/,$_))[2,3];
     $reftype[$refnr] =~ tr/A-Z/a-z/;

#     print "ref[$refnr]: <",$reftype[$refnr],"> URL: <",$url[$refnr],">\n";
     $refnr++;
 }
}

# reftype = ReferenceNode->system, -> reference_system-> ref_system_name
# url = ReferenceNode->id
# reference_system(ref_system_id,ref_system_name)=($refnr, $reftype)
# reference ($ref_id,ref_system_id, ref_tag) = ($refnr,$reftype,$url)
close(REFFILE);

#
# First we check if classification is already in database
#
for $i (1..$classnr-1) 
{
   #print "$i: $classshort[$i] -> $classdesc[$i], Priority: $prio[$i] \n";

   $sql="SELECT sig_class_name FROM sig_class WHERE sig_class_id = $i";
   $sth=$dbh->prepare($sql);
   $sth->execute();
   $sig_class_name=$sth->fetchrow_array();
   $sth->finish();

   if (length ($sig_class_name)==0)
   {
#
# classification not yet in database
#
      $sql="INSERT INTO sig_class (sig_class_id, sig_class_name) VALUES ( $i, '$classdesc[$i]' )";
      $sth=$dbh->prepare($sql);
      $sth->execute();
      $sth->finish();
   }
   elsif ($sig_class_name ne $classdesc[$i])
  {
# 
# class name has changed
#
     $sql="UPDATE sig_class SET sig_class_name='$classdesc[$i]' WHERE sig_class_id = $i";
      $sth=$dbh->prepare($sql);
      $sth->execute();
      $sth->finish();
  }
}
#
# now we actualize reference_system
# 
for $i (0..$refnr-1) 
{
   $sql="SELECT ref_system_name FROM reference_system WHERE ref_system_id = $i";
   $sth=$dbh->prepare($sql);
   $sth->execute();
   $ref_system_name=$sth->fetchrow_array();
   $sth->finish();

   if (length ($ref_system_name)==0)
   {
#
# reference system not yet in database
#
      $sql="INSERT INTO reference_system (ref_system_id, ref_system_name) VALUES ( $i, '$reftype[$i]' )";
      $sth=$dbh->prepare($sql);
      $sth->execute();
      $sth->finish();
   }
   elsif ($sig_class_name ne $classdesc[$i])
  {
# 
# class name has changed
#
     $sql="UPDATE reference_system SET ref_system_name='$reftype[$i]' WHERE ref_system_id = $i";
      $sth=$dbh->prepare($sql);
      $sth->execute();
      $sth->finish();
  }
}

print "sig_class and reference_system are updated.\n";

#
# now we have to analyze the rule files, there we find the ref_tag for
# the reference table. Then we can fill in sig_reference and finally 
# we can insert the signature.
#
$AllRules=0;
$AllInserted=0;
$AllRefIns=0;
$AllReferences=0;
for $i (0..$#ARGV)
{
    print "Working on ",$ARGV[$i],"\n";
    open(RULE,"<$ARGV[$i]") or die "Can't open rules file $ARGV[$i]: $!";
    $rules=0;
    $inserted=0;
    $refinserted=0;
    $RefPerFile=0;
    while (<RULE>)
    {
	if (m/^alert /) 
	{
	    $rules++;
	    $AllRules++;
# blank old values
            $msg="";
	    $rev=0;
	    $class="";
	    $priority=0;
            $sid=0;
            $refcount=0;
# we have a rule
	    chop();
# find the otn
            $start=0;
            $end=0;
	    for $i (0..length($_))
	    {
		if (substr($_,$i,1) eq '(' && $start==0) { $start=$i ;}
		if (substr($_,$i,1) eq ')' ) { $end=$i ;}
	    }

            if ($end <=$start) 
	    {
		print "No OTN found -> skipping: $_\n";
		break;
	    }
		    
	    $otn= substr($_,$start+1,$end-$start-1);
#	    print "OTN: $otn\n";
	    
            $start=0;
	    $end=0;

	    for $i (0..length($otn))
	    {
		if (substr($otn,$i,1) eq ';') 
		{
		    if ( $i>0 && substr($otn,$i-1,1) eq '\\')
		    {
			# escaped ;-sign
			next;
		    }
		    $end=$i;
		    $key=substr($otn,$start,$end-$start);
		    $start=$end+1;
#		    print "Key: $key\n";
# decomposite key 
# remove leading spaces
		    while ( substr($key,0,1) eq ' ' ) 
		    {
			$key=substr($key,1,length($key)-1);
		    }
#		    print "Key: $key\n";

 		    for $j (0..length($key))
		    {
                       # if we find a \ then simply ignore 
                       # the next sign blindly.
                       if (substr($key,$j,1) eq '\\')
                         {
                           $j++;
                           next;
                         }

		       if (substr($key,$j,1) eq ':')
			{
			    $command=substr($key,0,$j);
			    $value=substr($key,$j+1);
			    break;
			}
		    }
#		    print "Command: $command, Value  : $value\n";
		    if ( $command =~ /^sid/  ) {
			$sid=$value;
			next;
		    }
		    if ( $command =~ /^rev/ ) {
			$rev=$value;
			next;
		    }
		    if ( $command =~ /^classtype/ ) {
			$class=$value;
			next;
		    }
		    if ( $command =~ /^priority/ ) {
			$priority=$value;
			next;
		    }
		    if ( $command =~ /^msg/ ) {
			$msg=$value;
			next;
		    }
		    if ( $command =~ /^reference/ ) {
			$reference[$refcount]=$value;
			$refcount++;
			$RefPerFile++;
	                $AllReferences++;
			next;
		    }
			
		}
	    } # otn
#	    print "\nrule decomposition:\n";
#	    print " sid  : $sid\n";
# remove leading spaces
	    while ( substr($msg,0,1) eq ' ' ) 
	    {
		$msg=substr($msg,1,length($msg)-1);
	    }

	    while ( substr($msg,length($msg)-1,1) eq ' ' ) 
	    {
		$msg=substr($msg,0,length($msg)-1);
	    }

	    if (substr($msg,0,1)='"')
	    {
		$msg=substr($msg,1,length($msg)-2);
	    }
#	    print " msg  : $msg\n";	    
# escaping characters in message
#
	    $msg=$dbh->quote($msg);
#	    print " msg  : $msg\n";
	    $class=~ tr/A-Z/a-z/;	    
	    while ( substr($class,0,1) eq ' ' ) 
	    {
		$class=substr($class,1,length($class)-1);
	    }

	    while ( substr($class,length($class)-1,1) eq ' ' ) 
	    {
		$class=substr($class,0,length($class)-1);
	    }

#	    print " class: $class\n";

            if ($class eq "")
            {
              print "No classification given for $msg -> assigning 'bad-unknown'\n";
              $class="bad-unknown";
            }

            $saveclass=$class;

	    for $j (1..$classnr-1) 
	    {
#		print "class: $classshort[$j] $j\n";
		if ($classshort[$j] eq $class)
		{
#		    print "Match!\n";
		    $class=$j; 
		    break;
		}
	    }
#	    print " class: $class\n";
	    if (($priority==0) && ($class ne $saveclass))
	    {
		$priority=$prio[$class];
#		print " prio set via classification\n";
	    }
            else
            {
              $priority=2;
              print " attempted to set the priortiy via classification, but couldn't find classifcation priority -> assigning priority=2\n";
            }
#	    print " prio : $priority\n";
#	    print " rev  : $rev\n";
#	    for $j (0..$refcount-1)
#	    {
#		print " ref$j: $reference[$j]\n";
#	    }
#
# now we have the rule separated now we can try to insert all values
# into the database (or update the already inserted?)
#
# First check for the signature
#
            #ensure the loop is at least one turn.
	    if ($PrioRange<1) {$PrioRange=1;}
	    for $P (0..$PrioRange-1)
	    {
		if ($PrioRange>1) 
		{ 
		    $priority=$P+1; 
#		    print "Priority set to: <$priority>\n";
		}
	    $sql="SELECT sig_id FROM signature WHERE sig_name = $msg AND sig_class_id=$class AND sig_priority=$priority AND sig_rev=$rev AND sig_sid=$sid";
#	    print "SQL> $sql\n";
	    $sth=$dbh->prepare($sql);
	    $sth->execute();
	    $sig_id=$sth->fetchrow_array();
	    $sth->finish();
	    if (length ($sig_id)==0)
	    {
# new rule -> insert it
		$sql2="INSERT INTO signature (sig_name, sig_class_id, sig_priority, sig_rev, sig_sid) VALUES ($msg, $class, $priority, $rev, $sid)";
		$sth=$dbh->prepare($sql2);
		$sth->execute();
		$sth->finish();
		# now try to get sig_id again
		$sth=$dbh->prepare($sql);
		$sth->execute();
		$sig_id=$sth->fetchrow_array();
		$sth->finish();
		if (length ($sig_id)!=0) { $inserted++; $AllInserted++;}
	    }
#	    print "sig_id: $sig_id\n";
	    if (length ($sig_id)==0)
	    {
		print "Still no sig_id --> Skipping rule\n";
		next;
	    }
#
# now we have the sig_sid, next is to insert/update the references
#
	    for $j (0..$refcount-1)
	    {
#		print " ref$j: $reference[$j]\n";
		$key=$reference[$j];
		$key=~ tr/A-Z/a-z/;	    
		for $k (0..length($key))
		{
		    if (substr($key,$k,1) eq ',')
		    {
			$ref_system_name=substr($key,0,$k);
			$ref_tag=substr($key,$k+1);
			break;
		    }
		}
# remove leading and tailing spaces
		while ( substr($ref_system_name,0,1) eq ' ' ) 
		{
		    $ref_system_name=substr($ref_system_name,1,length($ref_system_name)-1);
		}

		while ( substr($ref_system_name,length($ref_system_name)-1,1) eq ' ' ) 
		{
		    $ref_system_name=substr($ref_system_name,0,length($ref_system_name)-1);
		}

		$ref_system_id=-1;
		for $k (0..$refnr)
		{
		    if ($ref_system_name eq $reftype[$k])
		    {
			$ref_system_id=$k;
			break;
		    }
		}
		if ($ref_system_id==-1)
		{
		    print "Unknown reference_system: $ref_system_name -> Skipping \n";
		    next;
		}
#		print "ref_system_name: <$ref_system_name>\n";
#		print "ref_system_id  : <$ref_system_id>\n";
#		print "ref_tag        : <$ref_tag>\n";

		$sql="SELECT ref_id FROM reference WHERE ref_system_id = $ref_system_id AND ref_tag='$ref_tag'";
#	    print "SQL> $sql\n";
		$sth=$dbh->prepare($sql);
		$sth->execute();
		$ref_id=$sth->fetchrow_array();
		$sth->finish();

		if (length ($ref_id)==0)
		{
# not in database yet, inserting it
		    $sql2="INSERT INTO reference (ref_system_id, ref_tag) VALUES ($ref_system_id, '$ref_tag')";
#	    print "SQL> $sql\n";
		    $sth=$dbh->prepare($sql2);
		    $sth->execute();
		    $sth->finish();
#
# now fetch ref_id again
#
		    $sth=$dbh->prepare($sql);
		    $sth->execute();
		    $ref_id=$sth->fetchrow_array();
		    $sth->finish();
		    if (length ($ref_id)!=0) {$AllRefIns++; $refinserted++;}
		}
# should never happen
 		if (length ($ref_id)==0)
		{
		    print "Still no ref_id -> Skipping reference.\n";
		    next;
		}
#		print "ref_id: $ref_id, sig_id: $sig_id \n";
#
# now we have a sig_id and a ref_id, now we can fill in sig_reference
# first we check if the sig_refence is already there.
		$sql="SELECT sig_id FROM sig_reference WHERE sig_id=$sig_id AND ref_id = $ref_id";
#	    print "SQL> $sql\n";
		$sth=$dbh->prepare($sql);
		$sth->execute();
		$sig_id2=$sth->fetchrow_array();
		$sth->finish();

		if (length ($sig_id2)==0)
		{
# not in database yet, inserting it
		    $sql2="INSERT INTO sig_reference (sig_id, ref_seq, ref_id) VALUES ($sig_id, $j, $ref_id)";
#	    print "SQL> $sql\n";
		    $sth=$dbh->prepare($sql2);
		    $sth->execute();
		    $sth->finish();
# test if really in database
		    $sth=$dbh->prepare($sql);
		    $sth->execute();
		    $sig_id2=$sth->fetchrow_array();
		    $sth->finish();
		}
		if (length ($sig_id2)==0)
		{
		    print "Problem to insert sig_id: $sig_id, ref_id: $ref_id\n";
		    next;
		}

	    } # $j -> refcount
	} # $P -> PrioRange
	} # /^alert/
    } # while <RULE>
    close(RULE);
    print "Number of rules     : $rules\n";
    print "Inserted rules      : $inserted (PrioRange=$PrioRange -> ",$inserted/$PrioRange,")\n";
    print "Number of references: $RefPerFile\n";
    print "Inserted references : $refinserted\n";
    print "Done with   ",$ARGV[$i],"\n";
} # loop over rule files
print "Total of rules           : $AllRules\n";
print "Total inserted           : $AllInserted (PrioRange=$PrioRange -> ",$AllInserted/$PrioRange,")\n";
print "Total numer of references: $AllReferences\n";
print "Total references inserted: $AllRefIns\n";
exit;


$dbh->disconnect();
