#!/usr/bin/perl
#
# simple perl script to feed an RRD database with
# statistics from snort via an unix domain socket
#
# it is assumed that the data is actualized every 
# 30 seconds. So choose this value with snort, e.g.:
# 
#       snort -Z 30 ...
#
# Copyright (C) Dirk Geschke <Dirk@geschke-online.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
use IO::Socket;
use IO::Handle;
use POSIX qw(:sys_wait_h); # WNOHANG
use Socket;
use RRDs;
#
# this socket is created
#
$UXSOCKADDR="/tmp/stats";
#
# first we unlink the socket, maybe there is an old one...
#
unlink($UXSOCKADDR);
#
# create the socket 
#
$sock = IO::Socket::UNIX->new( Local => $UXSOCKADDR, Type => SOCK_DGRAM) 
  or die "Can't bind to Unix Socket: $!\n";
  $sock->setsockopt(SOL_SOCKET, SO_RCVBUF, 65440);

print "Ready to accept conntections!\n";

$RRDrecv="recv.RRD"; 
$RRDdrop="drop.RRD"; 
$RRDtcp="tcp.RRD";
$RRDudp="udp.RRD";
$RRDicmp="icmp.RRD";
$RRDarp="arp.RRD";
$RRDother="other.RRD";
$RRDdiscard="discard.RRD";
$RRDalert="alert.RRD";
$RRDlog="log.RRD";

if (! -e $RRDrecv) 
  {
#
# no rrd database exist so we have to create them
#
     $CreateRRD=true;
  }
while (1) {
  $len=44;
  $sock->recv($input,$len);
  $TotalEvents++;
  
  @fields=unpack(" L L L L L L L L L L L",$input);
  #print "\n";
  #for ($i=0; $i<11; $i++)
  #   {
  #     print "Field[$i]: $fields[$i]\n";
  #   }
  if ($CreateRRD eq true) 
    {
# 
# create RRD databases 
#
      RRDs::create ("$RRDrecv", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDdrop", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDtcp", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDudp", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDicmp", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDarp", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDother", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDdiscard", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDalert", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
      RRDs::create ("$RRDlog", "--start", "$fields[0]", "--step", "30",
                    "DS:Statistics:GAUGE:61:0:U",
                    "RRA:AVERAGE:0.5:1:100",
                    "RRA:AVERAGE:0.5:10:24",
                    "RRA:AVERAGE:0.5:20:144");
       $CreateRRD=false;
     }
#
# update the databases with actual values
#
    RRDs::update ($RRDrecv, "$fields[0]:$fields[1]");
    RRDs::update ($RRDdrop, "$fields[0]:$fields[2]");
    RRDs::update ($RRDtcp, "$fields[0]:$fields[3]");
    RRDs::update ($RRDudp, "$fields[0]:$fields[4]");
    RRDs::update ($RRDicmp, "$fields[0]:$fields[5]");
    RRDs::update ($RRDarp, "$fields[0]:$fields[6]");
    RRDs::update ($RRDother, "$fields[0]:$fields[7]");
    RRDs::update ($RRDdiscard, "$fields[0]:$fields[8]");
    RRDs::update ($RRDalert, "$fields[0]:$fields[9]");
    RRDs::update ($RRDlog, "$fields[0]:$fields[10]");

    #print " Time $fields[0]:$fields[1]\n";
}
# 
# example for a command to create a picture
# -s mentions the starttime in seconds of the epoche!
#
#  rrdtool graph pic.gif -s 1043925600 --title "Received Packets" DEF:recv=Sensor.RRD:Statistics:AVERAGE LINE2:recv#FF0000 

