#include "FLoP_endian.h"
#include "flopdebug.h"
/*
 * Function: CheckEndian
 *
 * Purpose: check endian type of the computer
 *
 * Arguments: none
 *
 * Returns: LITTLE_ENDIAN or BIG_ENDIAN as an integer
 *
 * Author: DG
 *
 */

char 
CheckEndian()
{
  union 
  { 
    struct { u_int8_t b1, b2; } Byte;
    u_int16_t Word;
  } Endian;


Endian.Word=0x1234;

 if (Endian.Byte.b1 == 0x12)
   {

     DEBUGF(DebugOutput(DEBUG_ENDIAN,1,"CheckEndian> Big Endian\n"););
     
     return(DG_BIG_ENDIAN);
   }
 else
   {

     DEBUGF(DebugOutput(DEBUG_ENDIAN,1,"CheckEndian> Little Endian\n"););

     return(DG_LITTLE_ENDIAN);
   }
 
}

/*
 * Function: convert16
 *
 * Purpose: converts a 16 bit value from one endian 
 *          to the other endian form
 *
 * Arguments: 16 bit value to convert as a pointer 
 *            type of endian - equal or different
 *
 * Returns: LITTLE_ENDIAN or BIG_ENDIAN as an integer
 *
 * Author: DG
 *
 */

void 
convert16(u_int16_t *Conv,int endian)
{
  union { 
    struct {u_int8_t b1, b2; } Byte;
    u_int16_t Word;
  } tmp1,tmp2;

  if (endian != DG_DIFFERENT) return;

  tmp1.Word=*Conv;
  tmp2.Byte.b1=tmp1.Byte.b2;
  tmp2.Byte.b2=tmp1.Byte.b1;
  *Conv=tmp2.Word;
}

/*
 * Function: convert32
 *
 * Purpose: converts a 32 bit value from one endian 
 *          to the other endian form
 *
 * Arguments: 32 bit value to convert as a pointer 
 *            type of endian - equal or different
 *
 * Returns: LITTLE_ENDIAN or BIG_ENDIAN as an integer
 *
 * Author: DG
 *
 */


void 
convert32(u_int32_t *Conv, int endian)
{
  union { 
    struct {u_int8_t b1, b2, b3, b4; } Byte;
    u_int32_t Long;
  } tmp1,tmp2;

  if (endian != DG_DIFFERENT) return;

  tmp1.Long=*Conv;
  tmp2.Byte.b1=tmp1.Byte.b4;
  tmp2.Byte.b2=tmp1.Byte.b3;
  tmp2.Byte.b3=tmp1.Byte.b2;
  tmp2.Byte.b4=tmp1.Byte.b1;

  *Conv=tmp2.Long;
}
