/**************************************************************************/
/*                                                                        */
/* This program is intended to receive snort alert data via an unix       */
/* domain socket based either on a specific priority (alert) or if too    */
/* many alerts are buffered (drop)                                        */
/* Depending on the compile switch the program either compiles as the     */
/* program drop (-DDROP) or alert (-UDROP). The alerts are send via mail  */
/* to a list of recipients                                                */
/*                                                                        */
/* These programs are part of the FLoP project                            */
/*                                                                        */
/* Author: Dirk Geschke <Dirk@geschke-online.de>                          */
/*                                                                        */
/**************************************************************************/

/*
** Copyright (C) 2002-2004 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


static char *rcsid="$Id: alert.c,v 1.36 2004/09/06 21:44:00 geschke Exp $";

#include "FLoPconfig.h"
#undef PACKAGE
#undef VERSION
#undef PACKAGE
#undef VERSION
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION

#include "comm.h"
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <time.h>
#include <syslog.h>
#include <signal.h>
#include "Signal.h"
#include "alert.h"
#include "flopdebug.h"
#if defined HAVE_BIND_HEADER && defined LINUX
#include <bind/netdb.h>
#else
#include <netdb.h>
#endif

#ifndef HAVE_U_INT8_T 
#define u_int8_t  uint8_t
#endif
#ifndef HAVE_U_INT16_T
#define u_int16_t uint16_t
#endif
#ifndef HAVE_U_INT32_T
#define u_int32_t uint32_t
#endif


#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX   108
#endif


          /* Prototypes */

void ArgCheck(char *arg);
void ReadConfig();
int OpenSocket();
void CmdLine(int argc, char* argv[]); 
void Usage(char *argv);
char *Time();
void StoreLine(char *format,...);
int MailInit();
int ConnectServer(char *ServerName, u_int16_t Port);
int SendMail();
void SigAlarm();
void SigInt();
void SigHup();
void DumpConfig();
void FreeBody();
#ifdef DROP
void DumpStdout();
#endif
int getdaemon(char *dir, mode_t mask);

         /* Global values */

char ConfigFile[STRLEN];
#ifdef DROP
char SocketName[STRLEN]="/tmp/drop";
char PIDFile[STRLEN]="/tmp/drop.pid";
#else
char SocketName[STRLEN]="/tmp/alert";
char PIDFile[STRLEN]="/tmp/alert.pid";
#endif
u_int SocketNameSet=0;
Mail *start=NULL;
Mail *ToStart=NULL;
Mail *Rcpts;
u_int Alerts=0;
u_int AlarmDelay=5*60;
u_int AlarmDelaySet=0;
u_int AlarmLevel=0;
u_int AlarmLevelSet=0;

#ifdef DEBUGFLOP
u_int8_t DebugVector[DEBUG_MAX]={0};
#endif

char DaemonDir[STRLEN]="";
char MailRcpt[STRLEN]="root";
u_int MailRcptSet=0;
char MailSender[STRLEN]="snort@genua.de";
u_int MailSenderSet=0;
/* MailDomain is only for the HELO line needed */
char MailDomain[STRLEN]="genua.de";
u_int MailDomainSet=0;
char MailServer[STRLEN]="localhost";
u_int MailServerSet=0;
u_int16_t MailPort=25;
u_int MailPortSet=0;
u_int NrRcpts=0;
u_int DumpConf=0;
u_int FailCount=0;
u_int MaxCount=5;
u_int MaxCountSet=0;
u_int LogSyslog=0;
u_int Daemon=0;
u_int DaemonSet=0;
u_int DaemonDirSet=0;
u_int maskSet=0;
u_int PIDFileSet=0; 
mode_t mask=0;

/* Shall we lookup DNS names ?*/
/* Since the number of sensors will be small compared to the possible
   number of alerts we should think about gathering a list of addresses
   and resolve only names for new sensors -> to be done later? */
u_int FQNNames=0;

/* needed for restart */
char *prog;
char **args;
int sd;

int 
main(int argc, char* argv[])
{
  int i, size, length, found;
  AlertData alert;
  IP IP1,IP2;
  u_int8_t flags;
  char Flags[8];
  char line[STRLEN];
  char *ptr;
  char *hostname=NULL;
  pid_t pid;
  FILE *fd = NULL;

  struct hostent *hostinfo;
  struct servent *service;
  Sensor *FQNSensor=NULL;
  Sensor *FQNStart=NULL;
  Sensor *FQNLast;

  /* Store progname and arguments for restart */
  prog=argv[0];
  args=argv;

#ifdef DROP
   snprintf(ConfigFile,STRLEN,"%s/drop.conf",PATH_CONFIG);
#else
   snprintf(ConfigFile,STRLEN,"%s/alert.conf",PATH_CONFIG);
#endif
  if (ToStart==NULL)
    {
      /* Alloc first structure to hold the pointer to the messages */
      /* This is stored in 'start' to have an entry point */
      ToStart=calloc(sizeof(struct _Mail),sizeof(char));
      if (ToStart==NULL)
        {
	   Printf("Problem to allocate memory in startup -> Exiting!\n");
	   exit(1);
	}
      Rcpts=ToStart;
      NrRcpts=0;
    }

  /* Parse command line arguments, maybe we get a new config file... */
  CmdLine(argc,argv);

  /* Read config file */
  ReadConfig();

   /* Set the umask */
   umask(mask); 

  if (Daemon) 
    {
      /* Switch into daemon mode */
      Printf("starting in daemon mode.\n");
      if ( DaemonDir[0]=='\0') 
      {
         if (getcwd(DaemonDir,STRLEN)==NULL)
	 { 
           Printf("can't get working directory: %s\n",strerror(errno));
           exit(1);
	  }
      }
      if (getdaemon(DaemonDir,mask))
        {
           Printf("requested daemon mode failed: %s\n",strerror(errno));
	   exit(1);
        }
    }
  
  /* Check for existing PID file and if there is a running process      */
  /* with this pid. We don't check which process runs, only if one does */
  if ( (fd=fopen(PIDFile,"r")) ) 
    {
      if (fgets(line,STRLEN,fd))
	{
	  pid=(pid_t) atoi(line);
	  if ( !kill(pid,0) )
	    {
	      Printf("Program seems to be running with pid: %u\n",(u_int) pid);
	      Printf("if not: remove pid file: %s\n",PIDFile);
	      fclose(fd);
	      exit(1);
	    }
	}
      fclose(fd);
    }
  
  /* get actual pid */
  pid=getpid();

  /* Try to write the pid to the PIDFile, removing any old one */
  if ( (fd=fopen(PIDFile,"w")) )
    {
      fprintf(fd,"%u\n",(u_int) pid);
      fclose(fd);
    }
  else
    {
      /* We are not able to write the pid file */
      Printf("Can not write pid file: %s",PIDFile);
      exit(1); /* Maybe we simply should ignore this? */
    }

  /* if we should resolve sensor names from the IP addresses we
     try to store the vlues for further use in a linked list to
     minimize the CPU usage. Otherwise for each alert a DNS query
     hast to be performed. Thus we allocate here memory for a
     linked list of already resolved IP addresses                 */
  if (FQNNames) 
    {
      FQNSensor=calloc(sizeof(struct _Sensor),sizeof(char));
      if (FQNSensor==NULL)
        {
	  /* calloc fails on startup! This should really never happen...*/
	   Printf("Problem to allocate memory in startup -> Exiting!\n");
	   exit(1);
	}
      FQNStart=FQNSensor;
    } 

  if (NrRcpts==0)
    {
      Printf("You need at least one mail recipient!\n");
      exit(1);
    }

  /* Dump config values to stdout */
  if (DumpConf) DumpConfig();

  /* Open Unix Domain Socket to receive alerts */
  sd=OpenSocket();
  
  /* If we use a periodic check for alerts we add here the
     signal handler for alarm. This is highly advisable!   */
  if (AlarmDelay) 
    {
      Signal(SIGALRM,SigAlarm);
      alarm(AlarmDelay);
    }

  /* for a clean exit... */
  Signal(SIGINT,SigInt);
  Signal(SIGTERM,SigInt);
  /* to restart the program */
  Signal(SIGHUP,SigHup);

/* End of init, start loop */

  while (1)
    {
      size=sizeof(AlertData);
      length=recv(sd,(void *) &alert,size,MSG_WAITALL);
      if (length != size) 
	{
	  /* There seems to be something wrong with the client */
          /* Maybe the client closed the connection?           */
	  Printf("WARNING: Read size (%i) is different"
		 " to structure size (%i)\n",length,size);
	  Printf("Ignoring Entry\n");
	  break;  /* exit loop */
	}

      if (Alerts==0) StoreLine("==========================================="
			       "====================================");
      Alerts++;

      DEBUGF(DebugOutput(DEBUG_MAIN,1,"Got alert number: %u\n",Alerts););

      IP1.IPc=alert.sensorIP;
      errno=0;
      
      /* Should we try to resolve hostname of sensor reporting
         this alert? -> Time and CPU consuming! 
         Therefore we build our own cache of names, if we
         already resolved a name we take this value.
         NOTE: There is no expiration date!                     */
      if (FQNNames)
	{
	  found=0;
	  FQNSensor=FQNStart;
	  FQNLast=FQNStart;
	  do 
	    {
	      if (FQNSensor->IP==alert.sensorIP)
	        {
	          found=1;
	          hostname=FQNSensor->Name;   
	          break;
	        }
	      FQNLast=FQNSensor;
	      FQNSensor=FQNSensor->next;
	    } while (FQNSensor != NULL);

	  /* We did not found a valid entry -> we allocate memory
             for a new structure and try to resolve the name */
	  if (!found)
	    {
             /* Start with empty entry */
              hostname=NULL;
	     
              FQNSensor=calloc(sizeof(struct _Sensor),sizeof(char));

	      /* if we don't get memory we should exit, or?   
	         But actually we simply ignore to resolve the 
		 name and instead disable resolution. Maybe with
                 the next alert we get some memory?              */
	      if (FQNSensor)
	        {
		  FQNLast->next=FQNSensor;
                  FQNSensor->IP=alert.sensorIP;

		  DEBUGF(DebugOutput(DEBUG_MAIN,6,
				     "Call of DNS resolver.\n"););

#ifdef HAVE_BIND_HEADER
                  hostinfo = getipnodebyaddr((const void * )&(alert.sensorIP),
	                	     sizeof(alert.sensorIP), AF_INET, &errno); 
#else                  
                  hostinfo = gethostbyaddr((const void * )&(alert.sensorIP),
                                             sizeof(alert.sensorIP), AF_INET);
#endif
		  if (hostinfo)
		    {
		      FQNSensor->Name=strdup(hostinfo->h_name);
		    }
		  /* if no hostinfo is available then the Name is due to
                     the calloc equals NULL */
		  hostname=FQNSensor->Name;

		  DEBUGF(DebugOutput(DEBUG_MAIN,1,"Sensor hostname: %s\n",
				     hostname););

#ifdef HAVE_BIND_HEADER
		  freehostent(hostinfo); 
#endif
		}
	    } /* end calloc test */
	} /* FQNNames */
      else
        {
	  hostname=NULL;
        }
      if ( hostname == NULL )
    /* we were not able to get the fqn */
	{
	  StoreLine("Sensor   : %u.%u.%u.%u",
		   IP1.IPb.b1, IP1.IPb.b2, IP1.IPb.b3, IP1.IPb.b4);
	}
      else
	{
          StoreLine("Sensor   : %s (%u.%u.%u.%u)",hostname,
		   IP1.IPb.b1, IP1.IPb.b2, IP1.IPb.b3, IP1.IPb.b4);
	}
      StoreLine("Alert ID : %u-%u",alert.sid,alert.cid);
      StoreLine("Generator: %u",alert.generator);
      StoreLine("Signature: %u",alert.sig_id);
      StoreLine("Reference: %u",alert.reference);

      if (alert.time) StoreLine("PktTime  : %s",ctime(&alert.time));
      else StoreLine("PktTime : not available");

      /* Shorten message to be less or equal 250 bytes */
      /* (should never happen to be a problem?) */
      alert.alertmsg[250]='\0';
      length=strlen(alert.alertmsg)+2;
      ptr=line;
      for (i=0; i<length; i++)  *(ptr+i)='-'; 
      *(ptr+length)='\0';
      StoreLine(" +%s+",line);
      StoreLine(" | %s |",alert.alertmsg);
      StoreLine(" +%s+\n",line);

      /* We should replace these values with the content */
      /* of the classification file? */
      StoreLine("Priority : %u",alert.priority);
      StoreLine("Class    : %u\n",alert.classification);

      IP1.IPc=alert.srcip;
      IP2.IPc=alert.dstip;

      switch (alert.protocol)
	{
        case IPPROTO_TCP: 
	  flags=alert.Flag.TCP.flags;
	  Flags[0]='\0';
	  if (flags & TH_SYN)  strcat(Flags,"S");
	  if (flags & TH_FIN)  strcat(Flags,"F");
	  if (flags & TH_ACK)  strcat(Flags,"A");
	  if (flags & TH_PUSH) strcat(Flags,"P");
	  if (flags & TH_RST)  strcat(Flags,"R");
	  if (flags & TH_URG)  strcat(Flags,"U");
	  if (flags & TH_RES1) strcat(Flags,"X");
	  if (flags & TH_RES2) strcat(Flags,"Y");
	  StoreLine("%s >>> TCP (%u) %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u %s\n",
		   Time(), alert.protocol, 
		   IP1.IPb.b1, IP1.IPb.b2, IP1.IPb.b3, IP1.IPb.b4, 
		   ntohs(alert.sport),
		   IP2.IPb.b1, IP2.IPb.b2, IP2.IPb.b3, IP2.IPb.b4, 
		   ntohs(alert.dport),Flags);
          service=getservbyport(alert.sport,"tcp");
	  if (service)
	    {
	      StoreLine("Source port: %s (%u)", service->s_name,
		       ntohs(alert.sport));
	    }
          service=getservbyport(alert.dport,"tcp");
	  if (service)
	    {
	      StoreLine("Destination port: %s (%u)", 
		       service->s_name, ntohs(alert.dport));
	    }
	  break;

	case IPPROTO_UDP:
	  StoreLine("%s >>> UDP (%u) %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",
		 Time(),alert.protocol,
		 IP1.IPb.b1, IP1.IPb.b2, IP1.IPb.b3, IP1.IPb.b4, alert.sport,
		 IP2.IPb.b1, IP2.IPb.b2, IP2.IPb.b3, IP2.IPb.b4, alert.dport);
	  break;
	  
	case IPPROTO_ICMP:
	  StoreLine("%s >>> ICMP (%u) %u.%u.%u.%u -> %u.%u.%u.%u\n",
		 Time(),alert.protocol,
		 IP1.IPb.b1, IP1.IPb.b2, IP1.IPb.b3, IP1.IPb.b4,
		 IP2.IPb.b1, IP2.IPb.b2, IP2.IPb.b3, IP2.IPb.b4);
	  StoreLine("ICMP Type: %u",alert.Flag.ICMP.ICMPCode);
	  StoreLine("ICMP Code: %u",alert.Flag.ICMP.ICMPType);
	  break;
	  
	default:
	  StoreLine("%s >>> (%u) %u.%u.%u.%u -> %u.%u.%u.%u\n", Time(),
		    alert.protocol,
		    IP1.IPb.b1, IP1.IPb.b2, IP1.IPb.b3, IP1.IPb.b4,
		    IP2.IPb.b1, IP2.IPb.b2, IP2.IPb.b3, IP2.IPb.b4);
	}
      StoreLine("===================================================="
		"===========================");
      if (AlarmLevel && (Alerts >= AlarmLevel)) 
	{

	  DEBUGF(DebugOutput(DEBUG_MAIN,1,
			     "Number of alerts reached AlarmLevel\n"););
	  DEBUGF(DebugOutput(DEBUG_MAIN,1,
			     "Sending alerts via mail to purge memory\n"););

	  if ( SendMail() )
	     {
		 /* Sending of Mail failed. */ 
		 FailCount++;
		 if (FailCount > MaxCount)
		 {
		     Printf("Unable to send Mail, MaxCount (%u) reached.\n",MaxCount);
		     Printf("Number of Alerts in buffer: %u\n",Alerts);
#ifdef DROP
		     /* If we are still unable to send a mail we
			simply dump the alerts to stdout and go
                        on with the procedure. */
		     Printf("Dumping mails to standard out.\n\n");
		     DumpStdout();
#else
		     /* If we want to send additional alerts it won't 
                        make much sense to dump it to stdout. We have 
			a real problem with the mailserver. So we 
			simply exit and hope that another process
                        will inform an admin about the problem! */
		     Printf("Exiting.\n");
		     exit(1);
#endif
		 }
	     }
	}
    }

  /* Ok, normally never reached but... */
  FQNSensor=FQNStart;
  while (FQNSensor->next)
    {
      FQNStart=FQNSensor;
      FQNSensor=FQNSensor->next;
      free(FQNStart->Name);
      free(FQNStart);
    }
  FQNSensor=FQNSensor->next;
  free(FQNSensor->Name);
  free(FQNSensor);

  unlink(SocketName);

  exit(0);
}


/* 
 * Function: ReadConfig(ConfigFile)
 *
 * Purpose: read configuration files to set at least the parameter
 *          for database access (avoid password on command line)
 *          spaces/tabs at the beginning are removed, spaces at
 *          the end too. But maybe we should use a better parser...
 *
 * Author: DG
 * 
 */

void 
ReadConfig()
{
  FILE *fd = NULL;
  char *token =NULL;
  char *arg=NULL;
  char *val=NULL;
  char line[STRLEN];
  int linenumber=0;
  int i,len;
#ifdef DEBUGFLOP
  uint DebugArea,DebugLevel;
  char *c;
#endif

  if ( (fd=fopen(ConfigFile,"r")) == NULL ) 
    {
       Printf("can not open config file \"%s\": %s\n",ConfigFile,
	      strerror(errno));
       exit(1);
    }
  while((fgets(line, STRLEN, fd))!=NULL ) 
    { 
      linenumber++;
      /* remove end of line */
      if ( (token = strchr (line,'\n')) ) *token='\0';
      /* skip if line starts with a comment: if no comment sign */
      /* is found we are sure that we can derefence the pointer */
      /* before any comment sign!*/

      if ( line[0]=='#' ) continue;

      /* remove any comments at end of line and if a comment */
      /* sign is escaped: remove backslash from string!      */
      token = line;
      while ( (token = strchr(token,'#')) && (*(token-1)=='\\' ) )
      {
         val=token;
         arg=token-1;
	 while (*arg) {
		 *arg=*token;
		 arg++;
		 token++;
	 }
	 token=val;
      }

      /* Maybe token is pointing now to a real comment sign */
      if (token && *token=='#') *token='\0';
      /* next if line empty (or only a comment) */
      if (line[0]=='\0') continue;
      /* Up to ":" or "=" we get the argument */
      if ( (strstr(line,":") == NULL) && (strstr(line,"=") == NULL) )
        {
           Printf("\nIgnoring unknown entry in config file: \n");
           Printf("\n> %s (missing : or = ?)\n\n",line);
           continue;
        }
      arg = strtok(line,":=");
      /* second argument is the value, remove tabs */
      val = strtok(NULL,"\0");
      /* remove leading spaces or tabs */
      while ( arg[0]==' ' || arg[0]=='\t') arg++;
      while ( val[0]==' ' || val[0]=='\t') val++;
      /* remove spaces at end of value */
      while ( (strlen(arg) > 1) && (arg[strlen(arg)-1] == ' ') )
		    arg[strlen(arg)-1]='\0';
      while ( (strlen(val) > 1) && (val[strlen(val)-1] == ' ') )
		    val[strlen(val)-1]='\0';

      DEBUGF(DebugOutput(DEBUG_READCFG,1,
			 "Line: %i Argument: <%s> Value: <%s>\n",
			 linenumber,arg,val););
      ArgCheck(val);

      len=strlen(val);
      if(len > STRLEN-2) 
	{
	  Printf("Value in config file too long: %s\n",val);
	  exit(1);
	}

      if ( !strncasecmp(arg,"MailRecipient",13) )
        {
	  strncpy(MailRcpt,val,len);
	  MailRcpt[len]='\0';

	  Rcpts->line=strdup(MailRcpt);
	  Rcpts->next=calloc(sizeof(struct _Mail),sizeof(char));
	  Rcpts=Rcpts->next;
	  NrRcpts++;

	  continue;
        } 

      if ( !strncasecmp(arg,"MailSender",10) )
        {
	  if(!MailSenderSet)
	    {
	      strncpy(MailSender,val,len);
	      MailSender[len]='\0';
	    }
	   continue;
        } 

      if ( !strncasecmp(arg,"MailDomain",10) )
        {
	  if(!MailDomainSet)
	    {
           strncpy(MailDomain,val,len);
	   MailDomain[len]='\0';
	    }
	   continue;
        } 

      if ( !strncasecmp(arg,"MailServer",10) )
        {
	  if(!MailServerSet)
	    {
	      strncpy(MailServer,val,len);
	      MailServer[len]='\0';
	    }
	   continue;
        } 

      if ( !strncasecmp(arg,"MailPort",9) )
        {
	  if (!MailPortSet)
	    {
	      MailPort=atoi(val);
	    }
	  if (MailPort==0) MailPort=25;
	  continue;
        } 

      if ( !strncasecmp(arg,"MaxCount",8) )
        {
	  if (!MaxCountSet)
	    {
	      i=atoi(val);
	      if (i>0) MaxCount=i;
	    }
	  continue;
        } 
      
      if ( !strncasecmp(arg,"PIDFile",7) )
        {
	  if(!PIDFileSet)
	    {
           strncpy(PIDFile,val,len);
	   PIDFile[len]='\0';
	    }
	   continue;
        } 

      if ( !strncasecmp(arg,"SocketName",10) )
        {
	  if(!SocketNameSet)
	    {
	      if (len > UNIX_PATH_MAX)
		{
		  Printf("Length of SocketName too large!\n");
		  Printf("Maximum length: %u\n",UNIX_PATH_MAX);
		  Printf("Actual length : %u\n",len);
		  Printf("Exiting.\n");
		  exit(1);
		}
	      strncpy(SocketName,val,len);
	      SocketName[len]='\0';
	    }
	   continue;
        }

      if ( !strncasecmp(arg,"AlarmDelay",10) )
	{
	  if (!AlarmDelaySet)
	    {
	      i=atoi(val);
	      if (i>0) AlarmDelay=atoi(val)*60;
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"AlarmLevel",10) )
	{
	  if (!AlarmLevelSet)
	    {
	      AlarmLevel=atoi(val);
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"FQNNames",8) )
	{
	  if (FQNNames==0)
	    {
	      FQNNames=atoi(val);
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"Syslog",6) )
	{
	  if (LogSyslog==0)
	    {
	      LogSyslog=atoi(val);
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"DaemonMode",10) )
	{
	  if (DaemonSet==0)
	    {
	      Daemon=atoi(val);
	      /* if daemon mode we have to use syslog */
	      if (Daemon) LogSyslog=1;
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"umask",5) )
	{
	  if (!maskSet)
	    {
	      i=strtol(val,NULL,0);
	      if ( i > 0) mask=i;
	      else mask=0;
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"DaemonDir",5) )
	{
	  if (!DaemonDirSet)
	    {
	      strncpy(DaemonDir,val,len);
	      DaemonDir[len]='\0';
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"DEBUG",5) )
	{
#ifdef DEBUGFLOP
	  c=strchr(val,',');
	  if (c == NULL)
	    {
	      DebugArea=atoi(val);
	      DebugLevel=9;
	    }
	  else
	    {
	      *c='\0';c++;
	      DebugArea=atoi(val);
	      DebugLevel=atoi(c);
	    }

	  /* DebugLevel has maximum of 9 */
	  if (DebugLevel > 9 ) DebugLevel=9;

          /* Set debug value for area */
	  if ( (DebugArea < DEBUG_MAX) && (DebugArea > 0) )
	    {	    
              if (DebugVector[DebugArea]==0)
		{
		  DebugVector[DebugArea]=DebugLevel;
		  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
				     "DebugArea: %u, DebugLevel: %u\n",
				     DebugArea,DebugLevel););
		}
	    }
	  
	  /* if area is ascii "ALL" then set all values ( if not 
           *  already set to an value) to DebugLevel 
	   */
	  if ( DebugArea == 0 && (!strncasecmp(val,"ALL",3)) )
	    {
	      DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
				 "DebugArea: ALL, DebugLevel: %u\n",
				 DebugLevel););
	      for (i=0;i<DEBUG_MAX;i++)
		{
		  if (DebugVector[i] == 0 ) DebugVector[i]=DebugLevel;
		}
	    }
#else
	  Printf("Debugging is disabled. Please reconfigure FLoP with "
		 "debugging enabled.\n");
#endif
	  continue;
	}

      Printf("\nWARNING: unknown option in line %i: %s\n",linenumber,arg);
      
    }
  fclose(fd);
}


/* 
 *  Function: ArgCheck
 *
 *  Purpose: Check for occurence of wrapper around the arguments
 *           in the config file. We remove ' and " if they are at
 *           the beginning and the end.
 *
 *  Arguments: the extracted value as string
 *
 *  Author: DG
 *
 */

void 
ArgCheck(char *arg)
{
  char Value[STRLEN];
  int len;
  
  len=strlen(arg);

  if (len < 2) return;
  DEBUGF(DebugOutput(DEBUG_READCFG,5,"ArgCheck> Arg: %s, Len: %i\n",arg,len););

  if ( ((index(arg,'"')==arg) && (rindex(arg,'"')==arg+len-1)) ||
       ((index(arg,'\'')==arg) && (rindex(arg,'\'')==arg+len-1)) )
   {
       strncpy(Value,arg+1,len-2);
       /* dont't forget to end the string... */
       Value[len-2]='\0';

       DEBUGF(DebugOutput(DEBUG_READCFG,5,"ArgCheck> Value: %s\n",Value););

       strcpy(arg,Value);
   }
  return;
}

/* 
 * Function: CmdLine
 *
 * Purpose: Set or modify some values needed for startup
 *
 * Arguments: Commandline argv and number of elements argc
 *
 * Returns: Nothing, we set only global values
 *
 * Author: DG
 *
 */

void 
CmdLine(int argc, char* argv[]) 
{
  int opt, size, i;

#ifdef DEBUGFLOP
  uint DebugArea,DebugLevel;
  char *c;
  char options[]="bc:Dhvs:lL:m:A:r:f:d:M:p:S:V:Fw:P:";
#else
  char options[]="bc:Dhvs:lL:m:A:r:f:d:M:p:S:Fw:P:";
#endif

  while((opt = getopt(argc, argv, options)) != -1) 
    {
      switch(opt)
	{
	case 'b':
	  Daemon=1;
	  LogSyslog=1;
	  DaemonSet=1;
	  break;

	case 'm':
	  mask=(mode_t) strtol(optarg,NULL,0);
	  maskSet=1;
	  break;

	case 'w':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(DaemonDir,optarg, size);
	  DaemonDir[size]='\0';
	  DaemonDirSet=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "Working directory for daemon mode: %s\n",
			     DaemonDir););
	  break;  

	case 'P':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(PIDFile,optarg, size);
	  PIDFile[size]='\0';
	  PIDFileSet=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "PIDFile: %s\n",PIDFile););
	  break;  

	case 'v':
#ifdef DROP
	  Printf("\ndrop server to create unix socket, read\n");
#else
	  Printf("\nalert server to create unix socket, read\n");
#endif
	  Printf("snort alerts from servsock and send e-mails \n");
	  Printf("to a list of users.\n");
	  Printf("\nWritten by Dirk Geschke\n\n");
	  Printf("%s\n\n",rcsid);
	  exit(0);
	  break;
      
	case 'h':
	  Usage(argv[0]);
	  break;

#ifdef DEBUGFLOP
	case 'V':
	  c=strchr(optarg,',');
	  if (c == NULL)
	    {
	      DebugArea=atoi(optarg);
	      DebugLevel=9;
	    }
	  else
	    {
	      *c='\0';c++;
	      DebugArea=atoi(optarg);
	      DebugLevel=atoi(c);
	    }

	  /* DebugLevel has maximum of 9 */
	  if (DebugLevel > 9 ) DebugLevel=9;

          /* Set debug value for area */
	  if ( (DebugArea < DEBUG_MAX) && (DebugArea > 0) )
	    {	    
	      DebugVector[DebugArea]=DebugLevel;
	      DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
				 "DebugArea: %u, DebugLevel: %u\n",
				 DebugArea,DebugLevel););
	    }
	  
	  /* if area is ascii "ALL" then set all values ( if not 
           *  already set to an value) to DebugLevel 
	   */
	  if ( DebugArea == 0 && (!strncasecmp(optarg,"ALL",3)) )
	    {
	      DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
				 "DebugArea: ALL, DebugLevel: %u\n",
				 DebugLevel););
	      for (i=0;i<DEBUG_MAX;i++)
		{
		  if (DebugVector[i] == 0 ) DebugVector[i]=DebugLevel;
		}
	    }
	  break;
#endif

	case 'c':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(ConfigFile,optarg, size);
	  ConfigFile[size]='\0';

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "ConfigFile: %s\n",ConfigFile););

	  break;  

	case 'D':
	  DumpConf=1;
	  break;

	case 's':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  if (size > UNIX_PATH_MAX)
	    {
	      Printf("Length of SocketName too large!\n");
	      Printf("Maximum length: %u\n",UNIX_PATH_MAX);
	      Printf("Actual length : %u\n",size);
	      Printf("Exiting.\n");
	      exit(1);
	    }
	  strncpy(SocketName,optarg, size);
	  SocketName[size]='\0';
	  SocketNameSet=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "Socket: %s\n",SocketName););
	  break;  
	  
	case 'r':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(MailRcpt,optarg, size);
	  MailRcpt[size]='\0';
	  Rcpts->line=strdup(MailRcpt);
	  Rcpts->next=calloc(sizeof(struct _Mail),sizeof(char));
	  Rcpts=Rcpts->next;
	  NrRcpts++;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "Mail Recipient: %s\n",MailRcpt););
	  break;  

	case 'f':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(MailSender,optarg, size);
	  MailSender[size]='\0';
	  MailSenderSet=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "Mail Sender: %s\n",MailSender););

	  break;  

	case 'F':
	  FQNNames=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"FQNNames is set\n"););
	  break;

	case 'd':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(MailDomain,optarg, size);
	  MailDomain[size]='\0';
	  MailDomainSet=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "Mail Domain: %s\n",MailDomain););
	  break;  

	case 'S':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(MailServer,optarg, size);
	  MailServer[size]='\0';
	  MailServerSet=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "Mail Server: %s\n",MailServer););
	  break;  

	case 'A':
	  i=atoi(optarg);
	  if (i>0) 
	    {
	      AlarmDelay=i*60;
	      AlarmDelaySet=1;
	    }

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "AlarmDelay: %u\n",AlarmDelay););
	  break;  

	case 'M':
	  i=atoi(optarg);
	  if (i>0) 
	    {
	      MaxCount=i;
	    }

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "MaxCount: %u\n",MaxCount););
	  break;

	case 'p':
	  i=atoi(optarg);
	  if (i>0) 
	    {
	      MailPort=i;
	      MailPortSet=1;
	    }
	  else
	    {
	      MailPort=25;
	    }

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "MailPort: %u\n",MailPort););
	  break;  

	case 'l':
	  LogSyslog=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "Logging to syslog\n"););
	  break;

	case 'L':
	  i=atoi(optarg);
	  if (i >= 0) 
	    { 
	      AlarmLevel=i;
	      AlarmLevelSet=1;
	    }

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "AlarmLevel: %u\n",AlarmLevel););
	  break;  
	  
	default:
	  Usage(argv[0]);
	}
    }
}


/* 
 * Function: OpenSocket 
 *
 * Purpose: opens unix socket in server mode
 *  
 * Arguments: name of socket to open
 *
 * Returns:filedescriptor of opened socket
 *
 * Author: DG
 */
 
int 
OpenSocket()
{
  int sock, length;
  struct sockaddr_un unix_addr;

  sock = socket(PF_UNIX, SOCK_DGRAM, 0) ;
  if (sock <0 ) 
    {
      Printf("OpenSocket> socket failed: %s\n", strerror(errno));
      exit(1);
    }

  bzero((char *) &unix_addr, sizeof(unix_addr));

  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path, SocketName);

  DEBUGF(DebugOutput(DEBUG_OS,1,
		     "OpenSocket> Socket: %s\n",unix_addr.sun_path););

  unlink(SocketName);
  length = sizeof(unix_addr.sun_path) + sizeof(unix_addr.sun_family);

  if (bind(sock, (struct sockaddr *) &unix_addr,length )) 
    {
      Printf("OpenSocket> bind failed: %s\n", strerror(errno));
      exit(0);
    }

  DEBUGF(DebugOutput(DEBUG_OS,1,
		     "OpenSocket> Socket created: %s!\n",SocketName););

  return sock;
}


/* 
 * Function: Usage
 *
 * Purpose: Display the possible command line arguments
 *
 * Arguments: argv, only needed to know the program name
 *
 * Returns: No return, simply exit...
 *
 * Author: DG
 *
 */

void 
Usage(char *argv)
{
  Printf("\nUsage: %s [-bDFhlpTv] [-c file] [-A time] [-d domain] [-f from] "
	 "[-L level] [-m mode] [-M max] [-p port] [-P PIDfile] [-r rcpt] "
	 "[-s socket] [-S server] "
#ifdef DEBUGFLOP
         "[-V area,level] "
#endif
	 "[-w dir] \n",argv);
  Printf("\n -A time       : delay in minutes to check for alert mails");
  Printf("\n -b            : start in daemon mode, go in background");
  Printf("\n -c config     : configuartion file");
  Printf("\n -d domain     : sending domain (HELO)");
  Printf("\n -D            : dump config to stdout");
  Printf("\n -f from       : who will be the sender");
  Printf("\n -F            : try to resolve sensor name via DNS");
  Printf("\n -h            : this information");
  Printf("\n -l            : use syslog instead of stdout");
  Printf("\n -L level      : level to send alert mails");
  Printf("\n -m mode       : umask value for daemon mode");
  Printf("\n -M maxcount   : maximum of allowed failures if mail is to be send");
  Printf("\n -p port       : port to use with mail server");
  Printf("\n -P PIDfile    : filename to store pid in");
  Printf("\n -r recipient  : who shall get the mails");
  Printf("\n -s socketname : name of unix socket to be created");
  Printf("\n -S mailserver : name of mail server to use");
  Printf("\n -v            : version information");
#ifdef DEBUGFLOP
  Printf("\n -V area,level : activates debugging, area could be 'ALL', "
	 "level up to 9");
#endif
  Printf("\n -w daemondir  : directory to change to in daemon mode");
  Printf("\n\n");
  exit(0);
}


/*
 * Function: Time()
 * 
 * Purpose: returns actual time in readable format without
 *          ending \n
 *
 * Author: DG
 *
 */

char *
Time()
{
  static char message[STRLEN];
  time_t timep;

  timep=time(NULL);
  strcpy(message,ctime(&timep));
  message[strlen(message)-1]='\0'; 
  return message;
}


/*
 * Function: StoreLine(format,...)
 * 
 * Purpose: stores line in preparation to send it via smtp
 *
 * Author: DG
 *
 */

void 
StoreLine(char *format,...)
{
  va_list va_args;
  char line[STRLEN]="";
  static Mail *Body;

  if (start==NULL)
    {
      /* Alloc first structure to hold the pointer to the messages */
      /* This is stored in 'start' to have an entry point */
      start=calloc(sizeof(struct _Mail),sizeof(char));
      Body=start;
    }

  va_start(va_args,format);
  vsnprintf(line,STRLEN,format,va_args);
  va_end(va_args);

  /* Allocate memory to store the message line */
  Body->line=malloc(strlen(line)+1);
  /* ...and then allocate memory for the next structure to hold later lines */
  Body->next=calloc(sizeof(struct _Mail),sizeof(char));
  strcpy(Body->line,line);
  
  /* Finally let Body point to the next structure */
  Body=Body->next;
}


/*
 * Function: MailInit(mailsession)
 * 
 * Purpose: Initializes smtp connection
 *
 * Author: DG
 *
 */

int 
MailInit()
{
  int sockfd;
  char buffer[STRLEN]="";
  char line[STRLEN]="";

  sockfd=ConnectServer(MailServer,MailPort);
  if (sockfd < 0)
    {
      Printf("Cannot connect mailserver %s:%u\n",MailServer,MailPort);
      return sockfd;
    }
  
  if (ReadLn(sockfd, buffer, STRLEN) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

  DEBUGF(DebugOutput(DEBUG_MAIL,1,"<<<< %s\n",buffer););

  if (strncmp(buffer,"220 ",4))
    {
      Printf("Unexpected server response: %s\n",buffer);
      return(-1);
    }

  snprintf(line,STRLEN,"HELO %s",MailDomain);
  if ( SendLn(sockfd, line, 250) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

  snprintf(line,STRLEN,"MAIL FROM: %s",MailSender);
  if ( SendLn(sockfd,line,250) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

  Rcpts=ToStart;
  while(Rcpts && Rcpts->next)
    {
      snprintf(line,STRLEN,"RCPT To: %s",Rcpts->line);
      if ( SendLn(sockfd,line,250) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }
      Rcpts=Rcpts->next;
    }

  snprintf(line,STRLEN,"DATA");
  if ( SendLn(sockfd, line, 354) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

  snprintf(line,STRLEN,"Date: %s",Time());
  if ( WriteLn(sockfd,line) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

#ifdef DROP
  snprintf(line,STRLEN,"Subject: %u Snort Alerts dropped",Alerts);
#else
  snprintf(line,STRLEN,"Subject: %u Snort Alerts",Alerts);
#endif

  if ( WriteLn(sockfd,line) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

  Rcpts=ToStart;
  while(Rcpts && Rcpts->next)
    {
      snprintf(line,STRLEN,"To: %s",Rcpts->line);
      if ( WriteLn(sockfd,line) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }
      Rcpts=Rcpts->next;
    }

  snprintf(line,STRLEN,"From: %s\n",MailSender);
  if ( WriteLn(sockfd,line) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

  return(sockfd);
}

/*
 * Function: SendMail()
 * 
 * Purpose: Sends Mail to server, working through whole list of messages
 *
 * Author: DG
 *
 */

int 
SendMail()
{
  Mail *Body;
  int sockfd;
  char buffer[STRLEN];

  /* Initialize smtp session */
  sockfd=MailInit();
  if (sockfd < 0 )
    {
      Printf("MailInit failed. Trying again later...\n");
      return(-1);
    }

  /* Loop through complete body */
  Body=start;
  while (Body && Body->next)
    {
      if ( WriteLn(sockfd,Body->line) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

      DEBUGF(DebugOutput(DEBUG_MAIL,1,">>>> %s\n",Body->line););

      Body=Body->next;
    }

  /* Rese start value */
  start=NULL;
  Alerts=0;

  /* close mail */
  snprintf(buffer,STRLEN,".");
  if ( SendLn(sockfd,buffer,250) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }
  snprintf(buffer,STRLEN,"QUIT");
  if ( SendLn(sockfd,buffer,221) )
      { 
	  /* Failed to send mail */
	  close(sockfd);
	  return(-1);
      }

  close(sockfd);

/* Ok, mail is sent, we can now free the memory */

  FreeBody();

  DEBUGF(DebugOutput(DEBUG_MAIL,1,"Mail sent.\n"););

  FailCount=0;
  return(0);
}


/*
 * Function: SigAlarm()
 * 
 * Purpose: Catch ALARM-Signal to send mails (if available)
 *
 * Author: DG
 *
 */

void 
SigAlarm()
{
  if (Alerts) SendMail();

  alarm(AlarmDelay);

  return;
}


/* 
 * Function: ConnectServer
 *
 * Purpose: opens TCP socket and connects to server
 *  
 * Arguments: name of (remote) server and port 
 *
 * Returns:filedescriptor of opened socket
 *
 * Author: DG
 */

int 
ConnectServer(char *ServerName, u_int16_t port)
{
#ifdef HAVE_BIND_HEADER
  int errnum;
#endif
  int h_addr_index;
  int sockfd;
  struct hostent *server_ptr;
  struct sockaddr_in srvin;
    
  bzero((void *)&srvin, sizeof(srvin));
  
  srvin.sin_port = htons((u_short) port);
  srvin.sin_family= AF_INET;

  DEBUGF(DebugOutput(DEBUG_CS,1,"CS> Port: %i (%i)\n",port,srvin.sin_port););

  /* get a socket */
  sockfd=socket(AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0 ) 
    {
      Printf("CS> socket failed: %s\n", strerror(errno));
      exit(1);
    }
  
  /* get address of server */

#ifdef HAVE_BIND_HEADER
  server_ptr=getipnodebyname(ServerName, AF_INET, AI_ALL, &errnum);
#else
  server_ptr=gethostbyname(ServerName);
#endif

  if (server_ptr == NULL) 
    {
      Printf("%s: unknown host\n",ServerName);
      exit(1);
    }

  DEBUGF(DebugOutput(DEBUG_CS,1,"CS> Hostname: %s\n",server_ptr->h_name););

  /* open connection to server, try all possible addresses */
  /* until first call to connect() gets a connection       */
  h_addr_index=0;
  while (server_ptr->h_addr_list[h_addr_index] != NULL)
    {
      memcpy((void *)&srvin.sin_addr, server_ptr->h_addr_list[h_addr_index], 
	     server_ptr->h_length);
      if (connect(sockfd,(struct sockaddr*) &srvin, sizeof(srvin)) < 0) 
	{
	  if (server_ptr->h_addr_list[++h_addr_index] != NULL) 
	    {
	      /* try next address, the last one failed and */
	      /* we still have addresses...                */
	      continue;
	    }
	  /* All possible addresses failed to connect */
	  Printf("CS> failed to connect to server: %s\n", strerror(errno));

#ifdef HAVE_BIND_HEADER
	  freehostent(server_ptr);
#endif

          /* return a negative value to inform the calling  */
	  /* function about the failed connection. This way */
          /* we can sleep and try it again at a later time  */ 
	  return(-1);
	}

      /* we found an adress which got a connect -> Leave loop */
      break; 
    }

#ifdef HAVE_BIND_HEADER
  freehostent(server_ptr); 
#endif

  DEBUGF(DebugOutput(DEBUG_CS,1,"CS> Connected to TCP server\n"););

  return sockfd;
}


/* 
 * Function: DumpConfig
 *
 * Purpose: Dump the actual configuration to stdout
 *  
 * Author: DG
 */

void 
DumpConfig()
{
    Printf("\n Configuration parameter:\n");
    Printf(" ========================\n");
    Printf(" ConfigFile : <%s>\n",ConfigFile);
    Printf(" PIDFile    : <%s>\n",PIDFile);
    Printf(" SocketName : <%s>\n",SocketName);
    Printf(" MailServer : <%s>\n",MailServer);
    Printf(" MailPort   : <%i>\n",MailPort);
    Printf(" Nr of Rcpts: %u\n",NrRcpts);
    Rcpts=ToStart;
    while(Rcpts && Rcpts->next)
      {
	Printf(" MailRcpt   : <%s>\n",Rcpts->line);
	Rcpts=Rcpts->next;
      }
    Printf(" MailSender : <%s>\n",MailSender);
    Printf(" MailDomain : <%s>\n",MailDomain);
    Printf(" AlarmDelay : <%i> (seconds)\n",AlarmDelay);
    Printf(" AlarmLevel : <%i>\n",AlarmLevel);
    if (FQNNames)
      {
	Printf(" FQNNames   : true\n");
      }
    else
      {
	Printf(" FQNNames   : false\n");
      }
    Printf("\n");
}


/* 
 * Function: SigInt
 *
 * Purpose: remove the old socket on exit (exit only via SIGINT!)
 *
 * Author: DG
 */

void 
SigInt()
{
#ifdef DROP
  DumpStdout();
#endif
  unlink(SocketName);
  unlink(PIDFile);
  exit(0);
}


#ifdef DROP
/* 
 * Function: DumpStdout
 *
 * Purpose: Dump all mails to stdout, needed if sending of mails fails
 *
 * Author: DG
 */

void 
DumpStdout()
{
  Mail *Body;

  /* Loop through complete body */
  Body=start;
  while (Body && Body->next)
    {
      Printf("%s\n",Body->line);
      Body=Body->next;
    }
  /* now we have to free the Mail struct */
  FreeBody();
  /* And finally give FailCount a new chance */
  FailCount=0;
}
#endif


/* 
 * Function: FreeBody
 *
 * Purpose: Frees all memory alloacted for the mail storage
 *
 * Author: DG
 */

void 
FreeBody()
{
  Mail *Body;
  Mail *last;

  Body=start;
  while (Body && Body->next)
    {
      last=Body->next;
      free(Body->line);
      free(Body);
      Body=last;
    }
  if (Body) 
    {
      free(Body->line);
      free(Body);
    }
}


/*
 *  Function: SigHup()
 *  
 *  Purpose: Catch HUP-Signal and do a restart
 * 
 *  Author: DG
 *
 */

void 
SigHup()
{
  sigset_t sigset,osigset;

  /* Disable alarm function */
  alarm(0);

  unlink(PIDFile);
  close(sd);
  unlink(SocketName);
  Printf("Restarting %s\n",prog);

  /* Reenable SIGHUP, we were called by signal handler for SIGHUP */
  /* so normally this signal is blocked during this routine and   */
  /* this behaviour is inherited by the execv-process             */
  sigaddset(&sigset,SIGHUP);
  sigprocmask(SIG_UNBLOCK,&sigset,&osigset);
  
  /* So finally restart the main program */
  execv(prog,args);

  /* never reached except if execv fails */
  exit(1); 
}
