#include "FLoPconfig.h"
#include "comm.h"
#include "flopdebug.h"

/* 
 * Function: SendLn(socket, line, expect)
 *
 * Purpose: sends a line to socket and reads the reply from it.
 *          the numerical part is compared with 'expect'
 *          Note: this is for communicating with a mail server
 *          Note: the first 3 characters of a mail server response
 *                are the numerical return code.
 *  
 * Arguments: socket filedescriptor, line to print and expected response
 *            value
 *
 * Author: DG
 */
int 
SendLn(int sockfd, char *line, int expect)
{
  char buffer[STRLEN]="";
  if ( WriteLn(sockfd, line) )
     {
	 return(-1);
     }

  DEBUGF(DebugOutput(DEBUG_UTIL,2,">>>> %s\n",line););

  if ( ReadLn(sockfd, buffer, STRLEN) )
     {
	 return(-1);
     }

  DEBUGF(DebugOutput(DEBUG_UTIL,2,"<<<< %s\n",buffer););

  if (atoi(buffer) != expect )
    {
      Printf("Unexpected response: %s\n",buffer);
      return(-1);
    }
  return(0);
}


/* 
 * Function: ReadLn(socket, line, length)
 *
 * Purpose: reads a line from socket of maximal length character
 *          Note: Final newline is replaced by EOL '\0'
 *  
 * Arguments: socket filedescriptor, line to print and maximum length
 *
 * Author: DG
 */

int 
ReadLn(int fd, char *buffer, int length)
{
  int i;
  char c, *ptr;

  ptr=buffer;
  for (i=0; i<length; i++)
    {
      if (read(fd, &c, 1) == 1)
	{
	  if (c == '\n')
	    {
	      *ptr='\0';
	      return(0);
	    }
	  else
	    {
	      *ptr=c;
	      ptr++;
	    }
	}
      else
	{
	  Printf("ReadLn> partner closed connection: %s\n", strerror(errno));
	  return(-1);
	}
    }
  /* Ok, no newline but shall we return? */
  /* Probably we run into trouble next time we read in a response...*/
  Printf("WARNING: no newline received!\n");
  /* exit(1); */
  return(0);
}


/* 
 * Function: Writeln(socket, line)
 *
 * Purpose: writes a line to socket, the final '\n' is added
 *  
 * Arguments: socket filedescriptor and line to print
 *
 * Author: DG
 */

int 
WriteLn(int fd, char *buffer)
{
  int length;
  char fin='\n';

  length=strlen(buffer);
  
  if (write(fd, buffer, length) != length)
	{
	  Printf("WriteLn> write error: %s\n", strerror(errno));
	  Printf("WriteLn> <%s>\n", buffer);
          return(-1);
	};
  /* Append end of line */
  if (write(fd, &fin, 1) != 1)
    {
      Printf("WriteLn> write error: %s\n", strerror(errno));
      return(-1);
    };
  return(0);
}

