/* $Id: socket_database.c,v 1.41 2005/01/17 21:22:15 geschke Exp $ */
/* $Revision: 1.41 $ */
/* 
** This routines are for access to the database, either MySQL or Postgresql.
** Some parts are based on an early version of spo_database.c of the snort
** project. So probably
**
** Portions Copyright (C) 2000,2001 Carnegie Mellon University
** Portions Copyright (C) 2001 Jed Pickel <jed@pickel.net>
**
** and the remaining parts are
**
** Copyright (C) 2002-2004 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

static char *rcsid="$Id: socket_database.c,v 1.41 2005/01/17 21:22:15 geschke Exp $";

#include "socket_database.h"
#include "flopdebug.h"

#define TCP_OFFSET_DG(tcph)        (((tcph).th_offx2 & 0xf0) >> 4)
#define TCP_X2_DG(tcph)    ((tcph).th_offx2 & 0x0f)
#define IP_VER_DG(iph)     (((iph).ip_verhl & 0xf0) >> 4)
#define IP_HLEN_DG(iph)    ((iph).ip_verhl & 0x0f)

#ifdef GENUA
#warning GENUA defined!
#endif

/* in util.c: we first try to allocate as much memory as */
/* possible on startup. Thus we omit later checking for */
/* free memory */
void util_alloc();

static char *select_ptr=NULL; /* Memory is allocted in DatabaseInit! */
static char *insert_ptr=NULL; /* Memory is allocted in DatabaseInit! */
static char *tbuffer=NULL;  /* Needed in Database() to store palyoad */
static char *pbuffer=NULL;  /* Needed in Database() to store palyoad */

static u_int32_t EventOffset;

extern u_int LogSyslog;

/*
 * Function: DatabaseInit(DatabaseData *)
 *
 * Purpose: Connects to database, sets/gets the sensor id/name/encoding
 *          and selects the starting cid from database. Should we change
 *          this to scheme v106?
 *
 * Arguments: args => ptr to database parameters
 *
 * Returns: 0 on success
 *
 */

int 
DatabaseInit(DatabaseData *data)
{
  u_int32_t cid;
  int result, DBsuccess=0;

  DEBUGF(DebugOutput(DEBUG_DB,9,"DatabaseInit> Begin.\n"););

  /* This should never happen! */
  if(!data->sensor_name)
    {
      Printf("Database> ERROR: There is no sensor_name defined\n");
      exit(1);
    }

  data->tz = GetLocalTimeZone();

  DEBUGF(DebugOutput(DEBUG_DB,8,"DatabaseInit> Allocating "
		     "memory for select and insert strings.\n"););

  /* allocate memory for configuration queries */
  if (select_ptr==NULL)
    {
      select_ptr = (char *) malloc(MAX_QUERY_LENGTH+1);
      insert_ptr = (char *) malloc(MAX_INSERT_LENGTH+1);
    }
  if ( (select_ptr==NULL) || (insert_ptr==NULL) )
	{
	  Printf("DatabaseInit: problem to allocate memory -> Exiting.\n");
	  return(1);
	}

  /* These values are needed for the Database() function */
  if (tbuffer==NULL)
    {
      tbuffer = (char *) malloc ((MAXBUFFSIZE*2)+1);
      pbuffer = (char *) malloc ((sizeof(struct pcap_pkthdr)*2)+1);
    }

  if ( (tbuffer==NULL) || (pbuffer==NULL))
    {
      Printf("Database: problem to allocate memory -> Exiting.\n");
      return(1);
    }

  /* now we allocate the memory for the util functions */
  util_alloc();

  DEBUGF(DebugOutput(DEBUG_DB,9,"DatabaseInit> Before connect.\n"););

  /* Connect to database */
  if (Connect(data))
    {
      return(1);
    }

  DEBUGF(DebugOutput(DEBUG_DB,9,"DatabaseInit> After connect.\n"););


  /* Get DB Version, needed to check if last_cid has to be set! */
  data->DBschema_version = CheckDBVersion(data);

  DEBUGF(DebugOutput(DEBUG_DB,7, "DatabaseInit> schema version = %d\n", 
		     data->DBschema_version););

  if (data->DBschema_version > 105 )
    {
      snprintf(insert_ptr, MAX_INSERT_LENGTH, 
	       "INSERT INTO sensor (hostname, detail, encoding, last_cid) "
	       "VALUES ('%s','1','%u','0')", 
	       data->sensor_name,  data->encoding);
    }
  else
    {
      snprintf(insert_ptr, MAX_INSERT_LENGTH, 
	       "INSERT INTO sensor (hostname, detail, encoding) "
	       "VALUES ('%s','1','%u')", 
	       data->sensor_name,  data->encoding);
    }

  snprintf(select_ptr, MAX_QUERY_LENGTH, 
	   "SELECT sid FROM sensor WHERE hostname = '%s' "
	   "AND encoding = '%u' AND filter IS NULL",
	   data->sensor_name,  data->encoding);

#ifdef MYSQL_DB
  if ( (data->DBtype==DB_MYSQL) && data->transaction) 
    {
      DEBUGF(DebugOutput(DEBUG_DB,7, 
			 "DatabaseInit> MySQL - Setting autocommit to 0\n");); 
      Insert("SET AUTOCOMMIT=0",data);
    }
#endif

   if (data->transaction) DBsuccess=Insert("BEGIN",data);

   DEBUGF(DebugOutput(DEBUG_DB,6,"DatabaseInit> Select: <%s>\n",select_ptr););

   data->sid = Select(select_ptr,data);


   DEBUGF(DebugOutput(DEBUG_DB,6,"DatabaseInit> querying sensor sid: <%s>\n",
		      select_ptr););
   DEBUGF(DebugOutput(DEBUG_DB,5,"DatabaseInit> sensor sid: <%i>\n",
		      data->sid););

  if(data->sid == 0)
    {
      /* This sensor is not in the database so we add him */
      result=Insert(insert_ptr,data);
      if (data->transaction && DBsuccess) DBsuccess=result;

      if (result==0)
	{
	  DEBUGF(DebugOutput(DEBUG_DB,6,"DatabaseInit> Insert failed: <%s>\n",
			     insert_ptr););
	}

      data->sid = Select(select_ptr,data);
      if(data->sid == 0)
        {
	  Printf("Database> Problem obtaining SENSOR ID (sid) from "
		 "%s->sensor\n", data->dbname);
	  /* Should we exit here? I think we have a strange problem */
          /* with the database when this happens at this point...   */
	  Disconnect(data);
	  return(1);
        }
    }

   DEBUGF(DebugOutput(DEBUG_DB,5,"DatabaseInit> sensor id = %u\n", 
		      data->sid););

  /* Set the cid by getting the max value off the database */
  snprintf(select_ptr, MAX_QUERY_LENGTH, 
	   "SELECT MAX(cid) FROM event WHERE sid = '%u'",
          data->sid);

  data->cid=Select(select_ptr,data);

  DEBUGF(DebugOutput(DEBUG_DB,5,"DatabaseInit> last cid = %lu\n", 
		      data->cid););

    /* Get the versioning information for the DB schema */

  if ( data->DBschema_version < 104 )
    {
      Printf("Database> The underlying database seems to be running"
	     " an older version of the DB schema.\n"
	     "          Please re-run the DB creation script (create_mysql)\n"
	     "          located in the contrib\\ directory.\n");
      if (data->transaction) Insert("ROLLBACK",data);
      Disconnect(data);
      return(2);
    }
  else if ( data->DBschema_version < LATEST_DB_SCHEMA_VERSION )
    {                
      Printf("Database> The database is using an older"
	     " version of the DB schema\n");
    }
  else if (data->DBschema_version > 105 )
    {
        /* Set the cid by getting the max value off the database */
      snprintf(select_ptr, MAX_QUERY_LENGTH, 
	       "SELECT last_cid FROM sensor WHERE sid = '%u'",
	       data->sid);

      cid = Select(select_ptr,data);      

      if ( data->cid > cid ) 
	{
	  /* on last exit the last_cid field was not updated! */
	  snprintf(insert_ptr, MAX_INSERT_LENGTH, 
		   "UPDATE sensor SET last_cid='%u' WHERE sid='%u'", 
		   data->cid, data->sid);
	  result=Insert(insert_ptr,data);

          if (data->transaction && DBsuccess) DBsuccess=result;

	  if (result==0)
	    {
	      DEBUGF(DebugOutput(DEBUG_DB,6,
				 "DatabaseInit> Insert failed: <%s>\n",
				 insert_ptr););
	    }
	} 
      else
	{
	  /* Actualize data->cid to value of last_cid */
	  data->cid = cid;
	}

      if (data->FullPayload)
        {
	  snprintf(select_ptr,MAX_QUERY_LENGTH,"SELECT full_payload FROM schema"); 
	  if (Select(select_ptr,data) == 0)
            {
	      /* Database is not able to store full payload.
               * Print a warning message and a hint how to
               * activate it. Additionally we have to disable
               * the feature and go on (or should we exit?)
	       */
	      data->FullPayload=0;
	      Printf("The database is not able to store the full "
		     "payload of the alerts although it is requested.\n");
	      Printf("To allow it, please extend the database with the "
		     "following SQL statements:\n\n");
              /* Needed for full header of the network packet */
	      Printf("ALTER TABLE data ADD COLUMN data_header TEXT;\n");
              /* pcap header, containing timestamp and length */
	      Printf("ALTER TABLE data ADD COLUMN pcap_header TEXT;\n");
              Printf("ALTER TABLE schema ADD COLUMN full_payload SMALLINT;\n");
              Printf("UPDATE schema SET full_payload=1;\n\n");
	    }
	} /* FullPayload */
    }

      /* Check for column reference in table event */
      if (data->Reference)
        {
         snprintf(select_ptr,MAX_QUERY_LENGTH,"SELECT reference FROM schema"); 
         if (Select(select_ptr,data) == 0)
            {
             /* Database is not able to store the event reference.
               * Print a warning message and a hint how to
               * activate it. Additionally we have to disable
               * the feature and go on (or should we exit?)
              */
             data->Reference=0;
             Printf("The database is not able to store the reference of the"
                    " alerts in the event table although it is requested.\n");
             Printf("To allow it, please extend the database with the "
                    "following SQL statements:\n\n");
              /* Needed for full header of the network packet */
              Printf("ALTER TABLE event ADD COLUMN reference TEXT;\n");
              Printf("ALTER TABLE schema ADD COLUMN reference SMALLINT;\n");
              Printf("UPDATE schema SET reference=1;\n\n");
           }
       } /* Reference */


  /* Increment by one: this is the next cid! */
  data->cid++;

  /* store last_cid to use it as an offset to event_id and 
   * and event reference. This way we can achieve an unique
   * relation between event_id and event_reference. This is
   * useful in conjunction with tagged packets.
   */
  EventOffset = data->cid;

      /* Check for column data_header in table data */
  if (data->transaction)
    {
      if (DBsuccess) 
	{ 
	  /* all inserts succeeded so we can commit */
	  result=Insert("COMMIT",data);
	}
      else
	{
	  /* at least one insert failed */
	  result=Insert("ROLLBACK",data);
	  /* With DatabaseInit only the sensor is added to the database */
	  /* if this already fails it makes no sense to continue...    */
	  Printf("DatabaseInit> There was an error with the database\n");
	  Disconnect(data);
	  return(1);
	}
    }
  return(0);
}


/*
 * Function: Database(AlertPacket *, char * msg, void *arg)
 *
 * Purpose: Insert data into the database
 *
 * Arguments: p   => pointer to the current Alertpacket data struct 
 *            msg => pointer to the signature (alert) message
 *            arg => pointer to the data associated with the database
 *            event => pointer to the event structure
 *            Packet_ptr => pointer to the sniffed packet
 *
 * Returns: void function
 *
 */

int 
Database(AlertpktDB *p, char *msg, void *arg, Event *event, void *Packet_ptr)
{
  DatabaseData *data = (DatabaseData *)arg;
  char *time, *tmp=NULL;
  char *sig_msg;
  int i, dsize, hsize, ins;
  unsigned int sig_id=0;
  int NewRule;  
  int result, DBsuccess=1;
  
  DEBUGF(DebugOutput(DEBUG_DB,3,"Database> AlertpktDB: %p\n",p););
#ifndef GENUA
  DEBUGF(DebugOutput(DEBUG_DB,3,"Database> Message   : %s\n",msg););
#endif
  DEBUGF(DebugOutput(DEBUG_DB,3,"Database> DBdata    : %p\n",arg););
  DEBUGF(DebugOutput(DEBUG_DB,3,"Database> Packet_ptr: %p\n",Packet_ptr););

  if(msg == NULL)
    {
      msg = "";
    }

  /*** Build the query for the Event Table ***/
  if(!(p->val & NOPACKET_STRUCT))
    {
      time = GetTimeStamp((struct timeval *)&p->pkth.ts, data->tz);
    }
  else
    {
      /* Ok, there is no sniffed packet with timestamp so we */
      /* use the actual time. This is not very accurate here */
      time = GetCurrentTimeStamp();
    }

  if (event->sig_generator != GENERATOR_SNORT_ENGINE) /*(event->sig_id < 101) */
    {
       /* This alert is created by a preprocessor: We replace   */
       /* sig_rev with the generator id to separate the sig_sid */
       /* of the different sources. (sig_sid are only unique    */
       /* within each preprocessor! And yes: It is a hack...)   */

	event->sig_rev = event->sig_generator;
    }           

  if (data->transaction) DBsuccess=Insert("BEGIN",data);

  NewRule=1; /* First we don't trust anything! */

  if (data->DBtrust)
    {
       /* We trust now that the sig_id is already included in the */
       /* database. We can therefore skip all the next lines... */    

      snprintf(select_ptr, MAX_QUERY_LENGTH,
	       "SELECT sig_id FROM signature WHERE sig_priority = '%u' AND "
	       " sig_rev = '%u' AND sig_sid = '%u' ",
	       event->priority, event->sig_rev, event->sig_id);

      sig_id=Select(select_ptr,data);

  DEBUGF(DebugOutput(DEBUG_DB,3,
		     "Database> Trusting Database, sig_id: %i\n",sig_id););
  DEBUGF(DebugOutput(DEBUG_DB,3,"Database> %s\n",select_ptr););

      if (sig_id > 0 ) 
	 {
	     /* Yes, we can trust the DB... */
             NewRule=0;
          }
       else 
	{
           Printf("Disabling DBtrust modus for this alert:\n");
	   Printf("[%i] %s\n\n",event->sig_id,msg);
        }
    }
  if (NewRule)
    {
#ifdef GENUA
      Printf("Houston, we have a problem!\n");
      Printf("Here is a rule which is not in the database!\n");
      exit(1);
#else
      /* 
       * Determine the ID # of the signature of this alert 
       *
       * Note: Here we still use the old syntax of spo_database.c
       *       Therfore values which are zero are assumed to be
       *       not set so we ask for NULL values. I think a better
       *       approach is to ask for zeroes instead of NULL values
       *       like we do in DBTrust.
       *
       * Note: We don't count check here for the priority! So if
       *       there is a rule with several different priorities
       *       we still take the first sig_id of the database!
       *       This is another point which we should fix? But
       *       DBTrust will do it correctly.
       */

      sig_msg=snort_escape_string(msg,data);

      if ( event->sig_rev == 0 ) 
        {
          if( event->sig_id == 0) 
	    {
	      
	      snprintf(select_ptr, MAX_QUERY_LENGTH, 
		       "SELECT sig_id FROM signature WHERE sig_name = '%s' AND"
		       " sig_rev is NULL AND sig_sid is NULL ", sig_msg);
	    }
          else 
	    {
	      snprintf(select_ptr, MAX_QUERY_LENGTH, 
		       "SELECT sig_id FROM signature WHERE sig_name = '%s' AND"
		       " sig_rev is NULL AND sig_sid = %u ", 
		       sig_msg, event->sig_id);
	    }
        }
      else
        {
          if( event->sig_id == 0)
	    {
	      snprintf(select_ptr, MAX_QUERY_LENGTH,
		       "SELECT sig_id FROM signature WHERE sig_name = '%s' AND"
		       " sig_rev = %u AND sig_sid is NULL ",
		       sig_msg, event->sig_rev);
	    }
          else
	    {
	      snprintf(select_ptr, MAX_QUERY_LENGTH,
		       "SELECT sig_id FROM signature WHERE sig_name = '%s' AND"
		       " sig_rev = %u AND sig_sid = %u AND sig_priority = %u",
		       sig_msg, event->sig_rev, event->sig_id, event->priority);
	    }
        }
    
      sig_id = Select(select_ptr, data);
    
      DEBUGF(DebugOutput(DEBUG_DB,1,"Database> Signature sig_id: %i\n",
			 sig_id););

      /* We should have another method to insert the complete rule into */
      /* the database. Here we can only insert minimalistic information */
      /* At this point we have no OTN to fill in the missing values     */

      if(sig_id == 0)
        {
      
         /* Some snort preprocessors generate alerts with rev=1, */
         /* class=0, pri=0. By inserting signatures with these   */
         /* fields as NULL, they cannot be found when we look    */
	 /* for them later. Eventually, the signature table gets */
         /* clogged with duplicates. <Alex.Butcher@bristol.ac.uk>*/

          snprintf(insert_ptr, MAX_INSERT_LENGTH, "INSERT INTO signature "
		   "(sig_name, sig_priority, sig_rev, sig_sid, sig_class_id) "
                   "VALUES ('%s',%u,%u,%u,%u)", sig_msg, event->priority, 
		   event->sig_rev, event->sig_id,event->classification);

          result=Insert(insert_ptr,data);

	  if (result==0)
	    {
	      DEBUGF(DebugOutput(DEBUG_DB,6,"Database> Insert failed: <%s>\n",
				 insert_ptr););
	    }

	  /* if we use transaction and up to now we succeeded
	   * with inserts then we set DBsuccess with the result
	   * of the insert. On success it remains 1, otherwise
	   * it is set to 0. 
           */

	  if (DBsuccess) DBsuccess=result;

          sig_id = Select(select_ptr,data);
          if(sig_id == 0)
	    {
	      /* We still have no sig_id: This should not happen! */
	      Printf("Database> Problem inserting a new signature '%s'\n", msg);
              /* As stated above: Should we exit here? */
	    }
        } /* end of first: if (sig_id==0) */

      /* If we have a method to fill in the above values via another */
      /* method we can remove this code up to here... see -> DBtrust */
#endif /* GENUA */
    } /* end of NewRule */

  if (data->Reference)
    {
      if (data->Reference==2)
       {
         /* Add lastcid to get unique references even if snort is 
          * restarted. Otherwise snort will start with an event_id
          * of zero. Thus we will have different packets with the
          * same event_reference.
          */
         event->event_reference += EventOffset;
       }

      snprintf(insert_ptr, MAX_INSERT_LENGTH,
	       "INSERT INTO event (sid,cid,signature,timestamp,reference) "
	       "VALUES ('%u', '%u', '%u', '%s', '%u')",
	       data->sid, data->cid, sig_id, time, event->event_reference);
    }
  else
    {
      snprintf(insert_ptr, MAX_INSERT_LENGTH,
	       "INSERT INTO event (sid,cid,signature,timestamp) VALUES "
	       "('%u', '%u', '%u', '%s')",
	       data->sid, data->cid, sig_id, time);
    } /* Reference */

  result=Insert(insert_ptr,data);
  if (DBsuccess) DBsuccess=result;

  if (result==0)
    {
      DEBUGF(DebugOutput(DEBUG_DB,6,"Database> Insert failed: <%s>\n",
			 insert_ptr););
    }

  DEBUGF(DebugOutput(DEBUG_DB,8,"Database> p: %p\n",p););
  DEBUGF(DebugOutput(DEBUG_DB,8,"Database> p->val: %i\n",p->val););

  if(!(p->val & NOPACKET_STRUCT)) 
    {
      /* We do not log fragments! They are assumed to be handled */
      /* by the fragment reassembly pre-processor                */
      if((!p->frag_flag) && p->iph.ip_verhl ) 
	{
	  /*
	   *    >>> ICMP <<<
	   */
	  if(p->iph.ip_proto == IPPROTO_ICMP)
	    {
	      /*** Build a query for the ICMP Header ***/
	      if(p->icmp_ext)
		{
		  snprintf(insert_ptr, MAX_INSERT_LENGTH, 
			   "INSERT INTO icmphdr (sid, cid, icmp_type, "
			   "icmp_code, icmp_csum, icmp_id, icmp_seq) "
			   "VALUES ('%u','%u','%u','%u','%u','%u','%u')",
			   data->sid, data->cid, p->icmp_type, p->icmp_code,
			   ntohs(p->icmp_csum), ntohs(p->icmp_extid), 
			   ntohs(p->icmp_extseqno));
		}
	      else
		{
		  snprintf(insert_ptr, MAX_INSERT_LENGTH, 
			   "INSERT INTO icmphdr (sid, cid, icmp_type, "
			   "icmp_code, icmp_csum) "
			   "VALUES ('%u','%u','%u','%u','%u')",
			   data->sid, data->cid, p->icmp_type, p->icmp_code,
			   ntohs(p->icmp_csum));
		}
          result=Insert(insert_ptr,data);
	  if (DBsuccess) DBsuccess=result;

	  DEBUGF(DebugOutput(DEBUG_DB,6,"Database> ICMP: %s\n",insert_ptr););
	  if (result==0)
	    {
	      DEBUGF(DebugOutput(DEBUG_DB,6,"Database> Insert failed: <%s>\n",
				 insert_ptr););
	    }

	    }
	  else if (p->iph.ip_proto == IPPROTO_TCP ) /* && p->tcph.th_seq) */
	    /*
	     *              >>> TCP <<<
	     */
            {
	      /*** Build a query for the TCP Header ***/
	      snprintf(insert_ptr, MAX_INSERT_LENGTH, 
		       "INSERT INTO tcphdr "
		       "(sid, cid, tcp_sport, tcp_dport, tcp_seq,"
		       " tcp_ack, tcp_off, tcp_res, tcp_flags, tcp_win,"
		       " tcp_csum, tcp_urp) "
		       "VALUES ('%u','%u','%u','%u','%lu','%lu','%u',"
		       "'%u','%u','%u','%u','%u')",
		       data->sid, data->cid, ntohs(p->tcph.th_sport), 
		       ntohs(p->tcph.th_dport), (u_long)ntohl(p->tcph.th_seq),
		       (u_long)ntohl(p->tcph.th_ack), TCP_OFFSET_DG(p->tcph), 
		       TCP_X2_DG(p->tcph), p->tcph.th_flags, 
		       ntohs(p->tcph.th_win), ntohs(p->tcph.th_sum),
		       ntohs(p->tcph.th_urp));
	      result=Insert(insert_ptr,data);
	      if (DBsuccess) DBsuccess=result;

	      DEBUGF(DebugOutput(DEBUG_DB,6,"Database> TCP: %s\n",
				 insert_ptr););
	      if (result==0)
		{
		  DEBUGF(DebugOutput(DEBUG_DB,6,
				     "Database> Insert failed: <%s>\n",
				     insert_ptr););
		}

                    /*** Build the query for TCP Options ***/
	      for(i=0; i < (int)(p->tcp_option_count); i++)
		{
		  if((data->encoding == ENCODING_HEX) || 
		     (data->encoding == ENCODING_ASCII))
		    {
		      tmp = Fasthex(p->tcp_options[i].offset+Packet_ptr, 
				    p->tcp_options[i].len); 
		    }
		  else
		    {
		      tmp = Base64(p->tcp_options[i].offset+Packet_ptr, 
				   p->tcp_options[i].len); 
		    }
		  snprintf(insert_ptr, MAX_INSERT_LENGTH, 
			   "INSERT INTO opt "
			   "(sid, cid, optid, opt_proto, opt_code, opt_len, "
			   "opt_data) "
			   "VALUES ('%u','%u','%u','%u','%u','%u','%s')",
			   data->sid, data->cid, i, 6, p->tcp_options[i].code,
			   p->tcp_options[i].len, tmp); 
		  result=Insert(insert_ptr,data);
		  if (DBsuccess) DBsuccess=result;

		  DEBUGF(DebugOutput(DEBUG_DB,6,"Database> TCP: %s\n",
				     insert_ptr); );
		  if (result==0)
		    {
		      DEBUGF(DebugOutput(DEBUG_DB,6,
					 "Database> Insert failed: <%s>\n",
					 insert_ptr););
		    }
		}
	    }
	  else if(p->iph.ip_proto == IPPROTO_UDP && p->udph.uh_len)
	    {
	      /*
	       *             >>> UDP <<<
	       */
	      /*** Build the query for the UDP Header ***/
	      
	      snprintf(insert_ptr, MAX_INSERT_LENGTH,
		       "INSERT INTO udphdr "
		       "(sid, cid, udp_sport, udp_dport, udp_len, udp_csum) "
		       "VALUES ('%u', '%u', '%u', '%u', '%u', '%u')",
		       data->sid, data->cid, ntohs(p->udph.uh_sport), 
		       ntohs(p->udph.uh_dport), ntohs(p->udph.uh_len),
		       ntohs(p->udph.uh_chk));
	      result=Insert(insert_ptr,data);
	      if (DBsuccess) DBsuccess=result;

	      DEBUGF(DebugOutput(DEBUG_DB,6,"Database> UDP: %s\n",
				 insert_ptr););
	      if (result==0)
		{
		  DEBUGF(DebugOutput(DEBUG_DB,6,
				     "Database> Insert failed: <%s>\n",
				     insert_ptr););
		}
	    }
	}   
      
      /*** Build the query for the IP Header ***/
      if ( p->iph.ip_verhl )
	{
	  /*
	   *    >>> IP <<<
	   */
	  
	  snprintf(insert_ptr, MAX_INSERT_LENGTH, 
		   "INSERT INTO iphdr "
		   "(sid, cid, ip_src, ip_dst, ip_ver,"
		   "ip_hlen, ip_tos, ip_len, ip_id, ip_flags, ip_off,"
		   "ip_ttl, ip_proto, ip_csum) "
		   "VALUES ('%u','%u','%lu',"
		   "'%lu','%u',"
		   "'%u','%u','%u','%u','%u','%u',"
		   "'%u','%u','%u')",
		   data->sid, data->cid, (u_long)ntohl(p->iph.ip_src.s_addr), 
		   (u_long)ntohl(p->iph.ip_dst.s_addr), 
		   IP_VER_DG(p->iph), IP_HLEN_DG(p->iph), 
		   p->iph.ip_tos, ntohs(p->iph.ip_len), ntohs(p->iph.ip_id), 
		   p->frag_flag, ntohs(p->frag_offset), p->iph.ip_ttl, 
		   p->iph.ip_proto, ntohs(p->iph.ip_csum));
	  result=Insert(insert_ptr,data);
	  if (DBsuccess) DBsuccess=result;
	}

      DEBUGF(DebugOutput(DEBUG_DB,6,"Database> IP: %s\n",insert_ptr););
   if (result==0)
     {
       DEBUGF(DebugOutput(DEBUG_DB,6,"Database> Insert failed: <%s>\n",
			  insert_ptr););
     }
      
      /*** Build querys for the IP Options ***/
      for(i=0 ; i < (int)(p->ip_option_count); i++)
	{
	  if(&p->ip_options[i])
	    {
	      if((data->encoding == ENCODING_HEX) || 
		 (data->encoding == ENCODING_ASCII))
		{
		  tmp = Fasthex(p->ip_options[i].offset+Packet_ptr, 
				p->ip_options[i].len); 
		}
	      else
		{
		  tmp = Base64(p->ip_options[i].offset+Packet_ptr, 
			       p->ip_options[i].len); 
		}
	      snprintf(insert_ptr, MAX_INSERT_LENGTH, 
		       "INSERT INTO opt "
		       "(sid,cid,optid,opt_proto,opt_code,opt_len,opt_data) "
		       "VALUES ('%u','%u','%u','%u','%u','%u','%s')",
		       data->sid, data->cid, i, 0, p->ip_options[i].code,
		       p->ip_options[i].len, tmp); 
	      result=Insert(insert_ptr,data);
	      if (DBsuccess) DBsuccess=result;

	      DEBUGF(DebugOutput(DEBUG_DB,6,"Database> IP: %s\n",insert_ptr););
	      if (result==0)
		{
		  DEBUGF(DebugOutput(DEBUG_DB,6,
				     "Database> Insert failed: <%s>\n",
				     insert_ptr););
		}
	    }
	}
    }

  DEBUGF(DebugOutput(DEBUG_DB,5,
		     "Database> After packet header: %s\n",insert_ptr););

        /*** Build query for the payload ***/
  /* dsize: size of payload without headers, p->pkth.caplen is     */
  /*        total of captured packet, p->data is pointer to the   */
  /*        payload and Packet_ptr is the pointer to the beginning */
  /*        of the captured packet inclusive headers!              */
  /* dsize = (int) (Packet_ptr - p->data) + p->pkth.caplen; */
  /* maybe the captured len (caplen) is longer than the real length */
  /* Some UUUU's from 100 MBit? */

  /* First check if payload exists... */
  if (p->data)
   {
      dsize = (int) (Packet_ptr - p->data) + p->pkth.len;
      hsize = (int) (p->data) - (int) (Packet_ptr); 
   }
  else
   {
      /* no payload of higher protocol */
      dsize = 0;
      /* the whole packet is consists only of headers */
      hsize = p->pkth.len;
   }

  if (hsize > MAXBUFFSIZE) 
    {
      Printf("Error: Header size too large -> reducing to MAXBUFFSIZE\n");
      hsize=MAXBUFFSIZE;
    }

  DEBUGF(DebugOutput(DEBUG_DB,7,
		     "Database> DataSize: %i, HeaderSize: %i, Packet_ptr: %p"
		     ", data: %p\n", dsize, hsize, Packet_ptr,p->data););
  DEBUGF(DebugOutput(DEBUG_DB,7,
		     "Database> packet length: %u, captured length: %u\n",
		     p->pkth.len, p->pkth.caplen););

  if (hsize)
    {
      if(data->encoding == ENCODING_BASE64)
	{
	      tmp  = Base64((u_char *)Packet_ptr, hsize);
              strcpy(tbuffer,tmp);
              tmp  = Base64((u_char *)&p->pkth,sizeof(struct pcap_pkthdr));
              strcpy(pbuffer,tmp);
	}
      else
	{
	      tmp  = Fasthex((u_char *) Packet_ptr, hsize);
              strcpy(tbuffer,tmp);
              tmp  = Fasthex((u_char *)&p->pkth,sizeof(struct pcap_pkthdr));

	      DEBUGF(DebugOutput(DEBUG_DB,7,"Database> pcap_pkthdr: %i\n",
				 sizeof(struct pcap_pkthdr)););
	      DEBUGF(DebugOutput(DEBUG_DB,7,"Database> pcap-len: %i\n",
				 p->pkth.caplen););

              strcpy(pbuffer,tmp);
	}
    }

  if (dsize)
    {
      if(data->encoding == ENCODING_BASE64)
	{
	  tmp  = Base64((u_char *)p->data, dsize);
	}
      else
	{
	  if(data->encoding == ENCODING_ASCII)
	    {
	      tmp = Ascii((u_char *)p->data, dsize);
	    }
	  else
	    {
	      tmp  = Fasthex((u_char *)p->data, dsize);
	    }
	}
    }
  
  ins=0;

  if (hsize && dsize)
    {
      ins=1;
      if (data->FullPayload)
	{
	  snprintf(insert_ptr, MAX_INSERT_LENGTH - 3, 
		   "INSERT INTO data "
		   "(sid,cid,pcap_header,data_header,data_payload) "
		   "VALUES ('%u','%u','%s','%s','%s",
		   data->sid, data->cid, pbuffer, tbuffer, tmp);
	}
      else
	{
	  snprintf(insert_ptr, MAX_INSERT_LENGTH - 3, 
		   "INSERT INTO data "
		   "(sid,cid,data_payload) "
		   "VALUES ('%u','%u','%s",
		   data->sid, data->cid, tmp);
	}
    }
  else if ( hsize && (dsize==0) )
    {
      if (data->FullPayload)
	{
	  ins=1;
	  snprintf(insert_ptr, MAX_INSERT_LENGTH - 3, 
		   "INSERT INTO data "
		   "(sid,cid,pcap_header,data_header) "
		   "VALUES ('%u','%u','%s','%s",
		   data->sid, data->cid, pbuffer, tbuffer);
	}
    }

  /* Do we have something to insert? */
  if (ins)
    {
      strcat(insert_ptr, "')");
      result=Insert(insert_ptr,data);
      if (DBsuccess) DBsuccess=result;

      DEBUGF(DebugOutput(DEBUG_DB,6,"Database> data: %s\n",insert_ptr);); 
      if (result==0)
	{
	  DEBUGF(DebugOutput(DEBUG_DB,6,"Database> Insert failed: <%s>\n",
			     insert_ptr););
	}
    }

  data->cid++;

  DEBUGF(DebugOutput(DEBUG_DB,4,"Database> Return, next cid: %i.\n",
		     data->cid););

  if (data->transaction)
    {
      if (DBsuccess) 
	{ 
	  /* all inserts succeeded so we can commit */
	  result=Insert("COMMIT",data);
	}
      else
	{
	  /* at least one insert failed */
	  result=Insert("ROLLBACK",data);
	}
    }
  return(DBsuccess);
}

/* Some of the code in this function is from the 
   mysql_real_escape_string() function distributed with mysql.

   Those portions of this function remain
   Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB

   We needed a more general case that was not MySQL specific so there
   were small modifications made to the mysql_real_escape_string() 
   function. */

char * snort_escape_string(char * from, DatabaseData * data)
{
  /* the only string we have to escape is the altermsg */
  static char to_start[ALERTMSG_LENGTH*2+1];
  char * to;
  char * end; 
  int from_length;

  from_length = (int)strlen(from);

  to = to_start;
  {
    for(end=from+from_length; from != end; from++)
      {
        switch(*from)
        {
          case 0:             /* Must be escaped for 'mysql' */
            *to++= '\\';
            *to++= '0';
            break;
          case '\n':              /* Must be escaped for logs */
            *to++= '\\';
            *to++= 'n';
            break;
          case '\r':
            *to++= '\\';
            *to++= 'r';
            break;
          case '\\':
            *to++= '\\';
            *to++= '\\';
            break;
          case '\'':
            *to++= '\\';
            *to++= '\'';
            break;
          case '"':               /* Better safe than sorry */
            *to++= '\\';
            *to++= '"';
            break;
          case '\032':            /* This gives problems on Win32 */
            *to++= '\\';
            *to++= 'Z';
            break;
          default:
            *to++= *from; 
        }
      }
  }
  *to=0;
  return(char *)to_start;
}


/* Function: CheckDBVersion(DatabaseData * data)
 *
 * Purpose: To determine the version number of the underlying DB schema
 *
 * Arguments: database information
 *
 * Returns: version number of the schema
 */

int 
CheckDBVersion(DatabaseData * data)
{
  int schema_version;

  snprintf(select_ptr, MAX_QUERY_LENGTH,  "SELECT vseq FROM schema");

  schema_version = Select(select_ptr,data);

  return schema_version;
}


/* Function: Insert(char * query, DatabaseData * data)
 *
 * Purpose: Database independent function for SQL inserts
 * 
 * Arguments: query (An SQL insert)
 *
 * Returns: 1 if successful, 0 if fail
 */

int 
Insert(char * query, DatabaseData * data)
{
    int result = 0;

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(!(mysql_query(data->m_sock,query)))
        {
            result = 1;
        }
        else
        {
            if(mysql_errno(data->m_sock))
            {
              Printf("Database: mysql_error: %s\nSQL=%s\n", 
                           mysql_error(data->m_sock), query);
            }
        }
      }
#endif /* MYSQL */

#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_result = PQexec(data->p_sock,query);
	if (PQresultStatus(data->p_result) == PGRES_COMMAND_OK)
	    {
	      result = 1;
	    }
	else
	  {
	    if (PQerrorMessage(data->p_sock))
	      Printf("Database: psql_error: %s\nSQL=%s\n",
		   PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */

    if(result)
      {
	DEBUGF(DebugOutput(DEBUG_DB,3,"Insert> (%s) executed\n", query););
      }
    else
      {
	DEBUGF(DebugOutput(DEBUG_DB,3,"Insert> (%s) failed\n", query););
      }
    
    return result;
}


/* Function: Select(char * query, DatabaeData * data)
 *
 * Purpose: Database independent function for SQL selects that 
 *          return a non zero int
 * 
 * Arguments: query (An SQL insert)
 *
 * Returns: result of query if successful, 0 if fail
 */

int 
Select(char * query, DatabaseData * data)
{
    int result = 0;

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(mysql_query(data->m_sock,query))
	  {
            result = 0;
	  }
        else
	  {
            if(!(data->m_result = mysql_use_result(data->m_sock)))
	      {
                result = 0;
	      }
            else
	      {
                if((data->m_row = mysql_fetch_row(data->m_result)))
		  {
                    if(data->m_row[0] != NULL)
		      {
                        result = atoi(data->m_row[0]);
		      }
		  }
	      }
            mysql_free_result(data->m_result);
	  }
        if(!result)
	  {
            if(mysql_errno(data->m_sock))
	      {
                Printf("Database: mysql_error: %s\n", 
		       mysql_error(data->m_sock));
	      }
	  }
      }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_result = PQexec(data->p_sock,query);
	if ( PQresultStatus(data->p_result) == PGRES_TUPLES_OK)
	  {
	    if (PQntuples(data->p_result))
	      result = atoi(PQgetvalue(data->p_result,0,0));

	    DEBUGF(DebugOutput(DEBUG_DB,3,"Select> Status: %i\n",result););

	  }
	else
	  {
	    result=0;
	    if (PQerrorMessage(data->p_sock))
	      Printf("Database: psql_error: %s\nSQL=%s\n",
		     PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */

    if(result)
      {
	DEBUGF(DebugOutput(DEBUG_DB,3,"Select> (%s) returned %u\n", 
			   query, result););
      }
    else
      {
	DEBUGF(DebugOutput(DEBUG_DB,3,"Select> (%s) failed\n", query););
      }

    return result;
}


/* Function: Connect(DatabaseData * data)
 *
 * Purpose: Database independent function to initiate a database 
 *          connection
 */

int 
Connect(DatabaseData * data)
{
#ifdef MYSQL_DB  
  int x; 
  if ( data->DBtype == DB_MYSQL) 
    {
      data->m_sock = mysql_init(NULL);
      if(data->m_sock == NULL)
	{
	  Printf("Database: Connection to database '%s' failed\n", 
		 data->dbname);
	  return(2);
	}
      x = 0;
      if( mysql_real_connect(data->m_sock,"localhost", data->user,
	       data->password, data->dbname, x, data->socket_name, 0) == 0)
	{
	  if(mysql_errno(data->m_sock))
	    {
	      Printf("Database: mysql_error: %s\n", mysql_error(data->m_sock));
	      return(1);
	    }
	  Printf("Database: Failed to logon to database '%s'\n", data->dbname);
	  return(1);
	}
    }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_sock=PQsetdbLogin(NULL, data->socket_name ,NULL,NULL,
				  data->dbname, data->user, data->password);
	if (PQstatus(data->p_sock) == CONNECTION_BAD)
	  {
	    PQfinish(data->p_sock);
	    if (PQerrorMessage(data->p_sock))
	      Printf("Database: psql_error: %s\n",
		     PQerrorMessage(data->p_sock));
	    Printf("Database: Failed to logon to database '%s'\n",
		   data->dbname);
	    return(1);
	  }
      }
#endif /* POSTGRES */
    return(0);
}

/* Function: Disconnect(DatabaseData * data)
 *
 * Purpose: Close connection to database and free 
 *          allocated memory.
 */

void 
Disconnect(DatabaseData * data)
{
  static int called=0;

  if (called) return;
  called++;

  DEBUGF(DebugOutput(DEBUG_DB,3,
		     "Disconnect> Closing connection to database \"%s\"\n", 
		     data->dbname););

  if (data)
    {
      if ( data->DBschema_version > 105 )
	{
	  /* on exit update last_cid to last cid!*/
	  snprintf(insert_ptr, MAX_INSERT_LENGTH, 
		   "UPDATE sensor SET last_cid='%u' WHERE sid='%u'",
		   data->cid, data->sid);
	  Insert(insert_ptr,data);
	}
#ifdef MYSQL_DB
      if((data->DBtype==DB_MYSQL) && data->m_sock ) 
	mysql_close(data->m_sock);
#endif 
#ifdef POSTGRES
      if ( (data->DBtype==DB_POSTGRES) && data->p_sock) 
	PQfinish(data->p_sock);
#endif
    }

  /* to be kindly we free some memory... (but exit() will do it too... */
  free(insert_ptr);
  free(select_ptr);
}
