/* $Id: sockserv.c,v 1.56 2004/10/09 20:24:12 geschke Exp $ */
/* $Revision: 1.56 $ */
/*
** This program is part of the FLoP project 
**
** Copyright (C) 2002-2004 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

static char *rcsid="$Id: sockserv.c,v 1.56 2004/10/09 20:24:12 geschke Exp $";

#include "FLoP_endian.h"
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <pthread.h>
#include "spo_alert_unixsock_db.h"
#include <unistd.h>
#include <stdarg.h>
#include <signal.h>
#include "Signal.h"
#include "comm.h"
#include "flopdebug.h"
#include "io.h"

#if defined HAVE_BIND_HEADER && defined LINUX
#include <bind/netdb.h>
#else
#include <netdb.h>
#endif

#ifndef HAVE_U_INT8_T 
#define u_int8_t  uint8_t
#endif
#ifndef HAVE_U_INT16_T
#define u_int16_t uint16_t
#endif
#ifndef HAVE_U_INT32_T
#define u_int32_t uint32_t
#endif

#include <errno.h>

#ifdef GENUA
#warning GENUA defined!
#endif

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX   108
#endif

#define DG_MAXBUF 131072

#define MINWATER 100
#define MINWATERDIFF 10

/* Maximum fail counter for send() call */
#define MAXFC 5
#if defined DROP
#include "alert.h"
#endif

/* TYPEDEFS */

typedef struct _DGPacket 
{
  u_int8_t *pkt_ptr;            /* pointer to the network packet */
  AlertpktDB alertpkt;          /* AlterpktDB from snort...     */
  struct _DGPacket *next;       /* pointer to the next packet  */
  struct _DGPacket *prev;       /* pointer to the previous packet  */
} DGPacket;


/* PROTOTYPES */

char *Fasthex(u_char *xdata, int length);
int OpenSocket(char * socketname);
int ConnectServer(int port);
int Connect(char *servername, int port);
void *SendToServer(void *ptr);
void *ReadFromSocket(void *ptr);
void Usage(char *argv);
void CmdLine(int argc, char *argv[]);
void SigUsr();
void SigInt();
void SigAlarm();
void SigHup();
sighandler_t Signal(int signr, sighandler_t sighandler);
int getdaemon(char *dir,mode_t mask);

#ifdef DROP
void OpenDropSocket(char *filename);
int drop2socket(void * data, int size);
void DropPackets(DGPacket *last);
#endif


/* Global data */
pthread_mutex_t mutex;
pthread_cond_t thr_cond;
DGPacket *data;
char SocketName[STRLEN]="/tmp/snort";
char ServerName[STRLEN]="10.200.200.1";
char DaemonDir[STRLEN]="";
char PIDFile[STRLEN]="/tmp/sockserv.pid";
u_int port=1234;
u_int RFScounter=0;
u_int STScounter=0;
u_int MaxTry=120;
u_int WaitTime=30;
u_int AlarmDelay=0;
u_int LogSyslog=0;
u_int Daemon=0;
mode_t mask=0;

/* Size of SO_RCVBUF of unix domain socket -> maximum atomic size */
static int RcvBuf;

#ifdef DEBUGFLOP
u_int8_t DebugVector[DEBUG_MAX]={0};
#endif

#ifdef DROP
u_int32_t SensorIP;
static int DropSocket=0;
static char DropSocketName[STRLEN]="/tmp/drop";
struct sockaddr_un drop_addr;
u_int HighWater=10000;
u_int LowWater=9900;
pthread_t RFSthread=0;
#endif
u_int DropCounter=0;
pthread_t MainThread=0;

/* needed for restart */
char *prog;
char **args;
int ServerSocket;
int UnixSocket;
pid_t pid;

int 
main(int argc, char* argv[])
{
  pthread_t thr_rfs, thr_sts;
  pthread_attr_t thr_attr;
  int sock1,sock2;
  char line[STRLEN];
  char errbuf[STRLEN];
  FILE *fd = NULL;
#ifdef DROP
  struct sockaddr_in clientin;
  socklen_t clientlen;

  bzero((void *)&clientin, sizeof(clientin));
#endif
  
  /* Store progname and arguments for restart */
  prog=argv[0];
  args=argv;

  /* First parse the command line arguments */
  CmdLine(argc,argv);

  DEBUGF(DebugOutput(DEBUG_MAIN,9,"Start of sockserv\n"););
#ifdef DROP
  if (DropSocket)
    {
      if (LowWater < MINWATER) 
	{ 
	  LowWater=MINWATER;
	  Printf("LowWater too small -> increasing to %i\n",LowWater);
	}
      if ( (HighWater-LowWater) < MINWATERDIFF )
	{
	  Printf("Water mark difference too low -> disabling drop feature!\n");
	  Printf("Increase difference to at least %i\n",MINWATERDIFF);
	  DropSocket=0;
	}
    }
  if (DropSocket) OpenDropSocket(DropSocketName);
#endif

  /* Set the umask */
  umask(mask); 

  if (Daemon) 
    {
      Printf("starting in daemon mode.\n");
      if ( DaemonDir[0]=='\0') 
      {
         if (getcwd(DaemonDir,STRLEN)==NULL)
	 { 
            DG_strerror_r(errno,errbuf,STRLEN);
	    Printf("can't get working directory: %s\n",errbuf);
            exit(1);
	  }
      }
      if (getdaemon(DaemonDir,mask))
        {
           DG_strerror_r(errno,errbuf,STRLEN);
           Printf("requested daemon mode failed: %s\n",errbuf);
	   exit(1);
        }
      /* Since daemon mode closes all file descriptors we
       * close the syslog and reopen it to avoid writing to
       * the first opened filedescriptor which would be the
       * connecting sockserv process
       */
      closelog();
      openlog(prog,LOG_PID,LOG_LOCAL0);
    }

  /* Check for existing PID file and if there is a running process      */
  /* with this pid. We don't check which process runs, only if one does */
  if ( (fd=fopen(PIDFile,"r")) ) 
    {
      if (fgets(line,STRLEN,fd))
	{
	  pid=(pid_t) atoi(line);
	  if ( !kill(pid,0) )
	    {
	      Printf("Program seems to be running with pid: %u\n",(u_int) pid);
	      Printf("if not: remove pid file: %s\n",PIDFile);
	      fclose(fd);
	      exit(1);
	    }
	}
      fclose(fd);
    }
  
  /* get actual pid */
  pid=getpid();

  /* Try to write the pid to the PIDFile, removing any old one */
  if ( (fd=fopen(PIDFile,"w")) )
    {
      fprintf(fd,"%u\n",(u_int) pid);
      fclose(fd);
    }
  else
    {
      /* We are not able to write the pid file */
      Printf("Can not write pid file: %s",PIDFile);
      exit(1); /* Maybe we simply should ignore this? */
    }

  /* install signal handler to recognize end of child process */
  Signal(SIGUSR1,SigUsr);  
  Signal(SIGINT,SigInt);  
  Signal(SIGTERM,SigInt);  
  Signal(SIGHUP,SigHup);  
  Signal(SIGPIPE,SIG_IGN);  

  /* if AlarmDelay is set: install timer routine to display */
  /* statistics about collected and delivered alerts with a */
  /* time interval of AlarmDelay seconds                    */
  if (AlarmDelay) 
    {
      Signal(SIGALRM,SigAlarm);
      alarm(AlarmDelay);
    }
  /* Allocate memory for the first packet */
  data = (DGPacket *) calloc(sizeof(DGPacket), sizeof(char));
  if (data == NULL) 
    {
      Printf("MAIN> ERROR: Unable to allocate memory!\n");
      exit(1);
    }

  DEBUGF(DebugOutput(DEBUG_MAIN,5,"allocated packet data at %p\n",data););

  pthread_mutex_init(&mutex,NULL);
  pthread_attr_init(&thr_attr);

  /* should be joinable by default but who knows?*/
  pthread_attr_setdetachstate(&thr_attr, PTHREAD_CREATE_JOINABLE);

  MainThread=pthread_self();

  /* Open unix socket so snort can write to it */
  sock1=OpenSocket(SocketName);
  UnixSocket=sock1;

  /* Start first thread: Reading snort packets via unix socket */
  errno=pthread_create(&thr_rfs,NULL,ReadFromSocket,(void *)&sock1);
  if (errno != 0)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error with ReadFromSocket thread: %s\n",errbuf);
      exit(1);
    }

  /* Open TCP socket to server */
  sock2=Connect(ServerName,port);
  ServerSocket=sock2;

#ifdef DROP
  /* Get the local IP address of the sensor which connects to the server */
  clientlen=sizeof(clientin);
  getsockname(sock2,(struct sockaddr*) &clientin, &clientlen);
  SensorIP = clientin.sin_addr.s_addr;

  DEBUGF(DebugOutput(DEBUG_MAIN,1,"SensorIP: %x\n",htonl(SensorIP)););

#endif /* DROP */

  /* Second thread: send the data via TCP to central server */
  errno=pthread_create(&thr_sts,NULL,SendToServer,(void *)&sock2);
  if (errno != 0)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error with SendToServer thread: %s\n",errbuf);
      exit(1);
    }

  pthread_attr_destroy(&thr_attr);

/* Wait for end of threads. Should never happen?*/
  pthread_join(thr_sts,NULL);

  DEBUGF(DebugOutput(DEBUG_MAIN,5,"Exit of SendToServer thread!\n"););

  pthread_join(thr_rfs,NULL);

  DEBUGF(DebugOutput(DEBUG_MAIN,5,"Exit of ReadFromSocket thread!\n"););

 /* Ideally this point is never reached... */
  close(sock1);
  close(sock2);
  unlink(SocketName);
  pthread_mutex_destroy(&mutex);
  pthread_exit(NULL);
  exit(0);
}


/*
 * Function: SendToServer
 *
 * Purpose: Thread to send the snort data to a central server where it
 *          can be inserted via a unix socket to the database server
 *          not completely implemented now
 *
 * Arguments: Pointer to the filedescriptor of the TCP socket?
 *
 * Author: DG
 *
 */

void *
SendToServer(void *ptr) 
{
  int sock;
  int size, length;
#ifdef DEBUGFLOP
  char* text;
#endif
  DGPacket *STSdata,*tmp;
  static char buffer[STRLEN];
  static char errbuf[STRLEN];
  static int fc1, fc2;

  fc1=0;
  fc2=0;
  sock= *(int *) ptr;
  STSdata=data;
  
  /* Start endless loop for sending data as long as we have such... */
  while(1)
    {

      DEBUGF(DebugOutput(DEBUG_STS,7,"STS> data      at %p\n",STSdata););
      DEBUGF(DebugOutput(DEBUG_STS,7,"STS> data next at %p\n",STSdata->next););

      if (STSdata->next == NULL) 
	{

      DEBUGF(DebugOutput(DEBUG_STS,3,"STS> Waiting for mutex %i\n",
			 STScounter););

	/* No data available, so wait for incoming packets */
	/* The thread ReadFromSocket will send a signal if */
        /* a new packet is available.                      */
	pthread_mutex_lock(&mutex);
	pthread_cond_wait(&thr_cond,&mutex);
	pthread_mutex_unlock(&mutex);
      }

#ifdef DEBUGFLOP
      text=Fasthex((u_char *)STSdata->pkt_ptr,STSdata->alertpkt.pkth.caplen);
      DEBUGF(DebugOutput(DEBUG_STS,3,"STS> Got mutex %i\n",STScounter););
      DEBUGF(DebugOutput(DEBUG_STS,3,"STS> Paketsize: %i\n",STSdata->alertpkt.pkth.caplen););
      DEBUGF(DebugOutput(DEBUG_STS,3,"STS> Content  : %s\n",text););
      DEBUGF(DebugOutput(DEBUG_STS,3,"STS> Event ID : %i\n",STSdata->alertpkt.event.event_id););
      DEBUGF(DebugOutput(DEBUG_STS,3,"STS> Event Ref: %i\n",STSdata->alertpkt.event.event_reference););
      DEBUGF(DebugOutput(DEBUG_STS,3,"STS> Priority : %i\n",STSdata->alertpkt.event.priority););
      free(text);
#endif

      if ((length=recv( sock, buffer, STRLEN, MSG_DONTWAIT)) == 0)
	{
	  /* we should never receive data from the server */
          /* except for the one case: the server closes   */
          /* connection. In this case we close the socket */
          /* and try to get a new connection              */
	  DEBUGF(DebugOutput(DEBUG_STS,7,"STS> Return of recv %i %s\n",
			     length,buffer););

	  DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("Server closed socket before alert: %s\n",errbuf);
	  close(sock);

	  /* Need to re-open socket to server... */
	  sock=Connect(ServerName,port);

	  continue;
	}

      size=sizeof(AlertpktDB);

      /* MSG_WAIALL is not supported with older glibc's: This is already */
      /* done in via the wrapper function DGsend */
      /* length=send(sock,(void *)&STSdata->alertpkt,size,MSG_WAITALL);*/
      length=DGsend(sock, (void *) &STSdata->alertpkt, size, 0 );
      if (length != size) 
	{
          DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("STS> Write Error alertpkt: %s\n",errbuf);
	  Printf("WARNING: Write size (%i) is different" 
		 " to structure size (%i)\n",length,size);
	  if (length == -1)
	    {
	      /* There was an error with the send() command 
	       * retry to send it MAXFC times. 
               * Note: If the connection was closed then 
               * the above recv will recognize it and
               * tries to reopen the connection
               */
	      if (fc1 < MAXFC)
		{
		  Printf("Retrying again\n");
		  fc1++;
		  continue;
		}
	      else
		{
		  /* permanent error -> do a clean exit */

		  Printf("Permanent error to send alertpkt - Exiting!\n");

		  /* 
		   * We can't send the data so we have a major problem. 
                   * Solution: we do a clean exit, drop all buffered alerts 
		   * if necessary (and DROP is activated). Therefore we send 
		   * ourselve a SIGINT.
		   */
		  kill(pid,SIGINT);
		  exit(1);
		}

	    }
	}
      fc1=0;
       
      /* Do we have a sniffed packet ? */
      if (!(STSdata->alertpkt.val & NOPACKET_STRUCT)) 
	{
      if ((length=recv( sock, buffer, STRLEN, MSG_DONTWAIT)) == 0)
	    {   
	  /* we should never receive data from the server */
          /* except for the one case: the server closes   */
          /* connection. In this case we close the socket */
          /* and try to get a new connection              */

	      DEBUGF(DebugOutput(DEBUG_STS,7,"STS> Return of recv %i %s\n",
				 length,buffer););

	      DG_strerror_r(errno,errbuf,STRLEN);
	      Printf("Server closed socket before payload: %s\n",errbuf);
	      close(sock);
	      /* Need to reopen socket to server... */
	      sock=Connect(ServerName,port);
	      continue;
	    }

	  size=STSdata->alertpkt.pkth.caplen;

	  /* Send data to server 
           * MSG_WAITALL is not supported within older glibc's: This is 
           * already done via the wrapper function DGsend 
           */

	  length=DGsend(sock, (void *) STSdata->pkt_ptr, size, 0 );
	  if (length != size)
            {
              DG_strerror_r(errno,errbuf,STRLEN);
	      Printf("STS> Write error payload: %s\n",errbuf);
              Printf("WARNING: Write size (%i) is different"
		     " to structure size (%i)\n",length,size);
	      if (length == -1)
		{
		  /* There was an error with the send() command 
		   * retry to send it MAXFC times. 
		   */
		  if (fc2 < MAXFC)
		    {
		      Printf("Retrying again in payload section\n");
		      fc2++;
		      continue;
		    }
		  else
		    {
		      /* permanent error -> do a clean exit */
		      Printf("Permanent error to send payload - Exiting!\n");
		      /* 
		       * We can't send the data so we have a major problem. 
		       * Solution: we do a clean exit, drop all buffered 
		       * alerts if necessary (and DROP is activated). 
		       * Therefore we send ourselve a SIGINT.
		       */
		  kill(pid,SIGINT);
		      exit(1);
		    }
		}
            }
	  fc2=0;
        } /* end payload */
    
      /* Keep the next pointer, after the free() call it would be lost */
      tmp=STSdata->next;
      free(STSdata->pkt_ptr);
      free(STSdata);

      DEBUGF(DebugOutput(DEBUG_STS,7,"STS> Data freed\n"););

      /* Restore next pointer as actual pointer */
      STSdata=tmp;
      STScounter++;

      DEBUGF(DebugOutput(DEBUG_STS,7,"End of thread: %i\n",STScounter););

    }
  pthread_exit(NULL);
}


/*
 * Function: ReadFromSocket
 *
 * Purpose: Thread for reading snort alerts via unix socket
 *         
 * Arguments: Pointer to the unix socket 
 * 
 * Returns: Should never return...
 *
 * Author: DG
 */

void *
ReadFromSocket(void *ptr)
{
  AlertpktDB alertpkt;
  static char buf[DG_MAXBUF];
  static char errbuf[STRLEN];
  int sock,length,size;
  u_int oldHW,oldLW;
  DGPacket *last;

#ifdef DEBUGFLOP
  char *text;
#endif

  last=data;
 
#ifdef DROP
  /* store thread number for drop on SIGINT */
  RFSthread=pthread_self();
#endif

  sock= *(int *) ptr;

  /* Start endless loop reading from unix socket */
  while(1) 
    {
      /* First packet must be an alertpkt */
      size=sizeof(AlertpktDB);

      length=DGrecv(sock,(void *) &alertpkt,size, MSG_WAITALL);
      if (length != size) 
	{
          DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("RFS> read problem: %s\n",errbuf);
	  Printf("RFS> WARNING: Read size (%i) is different"
		 " to structure size (%i)\n",length,size);
	  Printf("RFS> Ignoring entry!\n");
	  continue;
	}
      RFScounter++;

      DEBUGF(DebugOutput(DEBUG_RFS,3,"RFS> Alarm        : %i\n",RFScounter););
#ifndef GENUA
      DEBUGF(DebugOutput(DEBUG_RFS,3,"RFS> Alert-Message: %s\n",
			 alertpkt.alertmsg););
#endif
      DEBUGF(DebugOutput(DEBUG_RFS,3,"RFS> Event ID     : %i\n",
			 alertpkt.event.event_id););
      DEBUGF(DebugOutput(DEBUG_RFS,3,"RFS> EventPriority: %i\n",
			 alertpkt.event.priority););

      /* Do we have a sniffed packet? */
      if(!(alertpkt.val & NOPACKET_STRUCT)) 
	{
	  /* This is an alert with packet structure, not all have such a 
	     structure, e.g. portscans have only an alert message */
	  size=alertpkt.pkth.caplen;

          if (size > RcvBuf)
	    {
	      Printf("RFS> pcap packet size %i larger than receive buffer of"
		     "unix domain socket %i\n",size,RcvBuf);
	      Printf("RFS> Ignoring alert: %s\n",alertpkt.alertmsg);
	      continue;
	    }

	  length=DGrecv(sock,(void *) &buf,alertpkt.pkth.caplen,MSG_WAITALL);
	  if (length != size) 
	    {
              DG_strerror_r(errno,errbuf,STRLEN);
	      Printf("RFS> read problem: %s\n",errbuf);
	      Printf("RFS> WARNING: Read size (%i) is different"
		     " to structure size (%i)\n",length,size);
	      Printf("RFS> Ignoring entry!\n");
	      continue;
	    }

#ifdef DEBUGFLOP
	  text=Fasthex((u_char *)&buf,alertpkt.pkth.caplen);
	  DEBUGF(DebugOutput(DEBUG_RFS,3,"RFS> Paketsize: %i, read: %i\n",alertpkt.pkth.caplen,length););
	  DEBUGF(DebugOutput(DEBUG_RFS,3,"RFS> Content: %s\n",text););
	  free(text);
#endif
	  data->pkt_ptr = malloc(alertpkt.pkth.caplen);
          if (data->pkt_ptr == NULL)
	    {
	      Printf("RFS> failed to malloc memory!\n");
#ifdef DROP
	      /* We adjust the HighWater and LowWater mark to
               * fool the DropPackets() routine to drop some
               * buffered alerts to refree some memory
               */
	      Printf("RFS> Start to drop some alerts.\n");
	      oldHW=HighWater;
	      oldLW=LowWater;
              HighWater=(RFScounter-DropCounter-STScounter);
	      if (DropSocket)
		{
		  Printf("Setting HighWater (%i) to %i\n",oldHW,HighWater);
		  if (HighWater <= LowWater)
		    {
		      LowWater=HighWater-MINWATER;
		      if (LowWater<MINWATERDIFF) LowWater=MINWATERDIFF;
		      if (LowWater>= HighWater)
			{
			  Printf("Unable to set LowWater mark (%i) to "
                                 "acceptable value -> Exiting.\n",LowWater);
			  exit(1);
			}
		      Printf("Adjusting LowWater (%i) to %i\n",oldLW,LowWater);
		    }
		  DEBUGF(DebugOutput(DEBUG_DROP,3,"RFS> Buffered alerts: %i "
				     "(new HighWater: %i)\n",
				     (RFScounter-DropCounter-STScounter),
				     HighWater););
		  DropPackets(last);
		}
	      /* Reset water marks to old values */
	      LowWater=oldLW;
	      HighWater=oldHW;
	      DEBUGF(DebugOutput(DEBUG_DROP,3,"RFS> Reset Water marks: LW = %i"
				 " HW = %i\n",LowWater,HighWater););
#else
	      /* Ok, we can't drop, so we wait a little bit and
               * retry again -> poor man solution
               */
	      Printf("Sleeping 1 second, waiting for free memory\n");
	      sleep(1);
#endif      
	      data->pkt_ptr = malloc(alertpkt.pkth.caplen);
	      if (data->pkt_ptr == NULL)
		{
		  Printf("RFS> failed again to malloc memory -> Exiting!\n");
		  kill(pid,SIGINT);
		  exit(1);
		}
	    }
	  bcopy(&buf,data->pkt_ptr,alertpkt.pkth.caplen);
	}

      /* Now we have to set a mutex lock. This is to ensure that   */
      /* we do not use an incomplete structure in the other thread */
      pthread_mutex_lock(&mutex);

      data->next = (DGPacket *) calloc(sizeof(DGPacket), sizeof(char));
      if (data->next == NULL)
	{
	  Printf("RFS> data->next failed to calloc memory!\n");
#ifdef DROP
	      /* We adjust the HighWater and LowWater mark to
               * fool the DropPackets() routine to drop some
               * buffered alerts to refree some memory
               */
	  Printf("RFS> Start to drop some alerts.\n");
	  oldHW=HighWater;
	  oldLW=LowWater;
	  HighWater=(RFScounter-DropCounter-STScounter);
	  if (DropSocket)
	    {
	      Printf("Setting HighWater (%i) to %i\n",oldHW,HighWater);
	      if (HighWater <= LowWater)
		{
		  LowWater=HighWater-MINWATER;
		  if (LowWater<MINWATERDIFF) LowWater=MINWATERDIFF;
		  if (LowWater>= HighWater)
		    {
		      Printf("Unable to set LowWater mark (%i) to "
			     "acceptable value -> Exiting.\n",LowWater);
		      exit(1);
		    }
		  Printf("Adjusting LowWater (%i) to %i\n",oldLW,LowWater);
		}
	      DEBUGF(DebugOutput(DEBUG_DROP,3,"RFS> Buffered alerts: %i "
				 "(new HighWater: %i)\n",
				 (RFScounter-DropCounter-STScounter),
				 HighWater););
	      DropPackets(last);
	    }
	  /* Reset water marks to old values */
	  LowWater=oldLW;
	  HighWater=oldHW;
	  DEBUGF(DebugOutput(DEBUG_DROP,3,"RFS> Reset Water marks: LW = %i"
			     " HW = %i\n",LowWater,HighWater););
#else
	  /* Ok, we can't drop, so we wait a little bit and
	   * retry again -> poor man solution
	   */
	  Printf("Sleeping 1 second, waiting for free memory\n");
	  sleep(1);
#endif      
	  data->next = (DGPacket *) calloc(sizeof(DGPacket), sizeof(char));
	  if (data->next == NULL)
	    {
	      Printf("RFS> failed again to calloc memory -> Exiting!\n");
	      kill(pid,SIGINT);
	      exit(1);
	    }
	}

      last = data; /* Pointer to actual data becomes last in next data...*/
      data->alertpkt=alertpkt;
      data=data->next;
      data->prev=last;

#ifdef DROP
      if (DropSocket && ((RFScounter-DropCounter-STScounter) > HighWater))
	{
	  DEBUGF(DebugOutput(DEBUG_DROP,3,"RFS> Buffered alerts: %i (%i)\n",
			     (RFScounter-DropCounter-STScounter),HighWater););
	  DropPackets(last);
	 }
#endif /* DROP */

      DEBUGF(DebugOutput(DEBUG_RFS,7,"RFS> Data      at %p\n",data););
      DEBUGF(DebugOutput(DEBUG_RFS,7,"RFS> Data next at %p\n",data->next););
      DEBUGF(DebugOutput(DEBUG_RFS,7,"RFS> Send signal!\n"););

      /* Now we can inform the other thread that there is a new    */ 
      /* complete packet available which can be send to the server */
     pthread_cond_signal(&thr_cond);

     DEBUGF(DebugOutput(DEBUG_RFS,7,"RFS> Sent signal!\n"););

     pthread_mutex_unlock(&mutex);

     DEBUGF(DebugOutput(DEBUG_RFS,7,"RFS> After unlock!\n"););

    }
   /* To be kindly... This stage should never be reached */
  pthread_exit(NULL);
}


/* 
 * Function: OpenSocket 
 *
 * Purpose: opens unix socket in server mode
 *  
 * Arguments: name of socket to open
 *
 * Returns: filedescriptor of opened socket
 *
 * Author: DG
 */
 
int 
OpenSocket(char * SocketName)
{
  int sock, length, old;
  char errbuf[STRLEN];

  sock = socket(PF_UNIX, SOCK_DGRAM, 0) ;
  if (sock <0 ) 
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("OpenSocket> socket problem: %s\n",errbuf);
      exit(1);
    }

  bzero((char *) &unix_addr, sizeof(unix_addr));

  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path, SocketName);

  DEBUGF(DebugOutput(DEBUG_OS,1,"OpenSocket> Socket: %s\n",
		     unix_addr.sun_path););

  unlink(SocketName);
  length = sizeof(unix_addr.sun_path) + sizeof(unix_addr.sun_family);

  if (bind(sock, (struct sockaddr *) &unix_addr,length )) 
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("OpenSocket> bind problem: %s\n",errbuf);
      exit(0);
    }

  DEBUGF(DebugOutput(DEBUG_OS,7,"OpenSocket> bind successful\n"););

/*
 * set receive buffer of unix socket to maximum possible value 
 * modulo 1024 -> this is to avoid a buffer overrun on heavy
 * attacks and a low machine...
 */
 
  /* first get old value, maybe it is already greater... */

  getsockopt(sock, SOL_SOCKET, SO_RCVBUF, &RcvBuf, &old);

  old=RcvBuf;

  DEBUGF(DebugOutput(DEBUG_OS,5,
		     "OpenSocket> start to increase receive buffer\n"););

  while ( RcvBuf<DG_MAXBUF && 
	  setsockopt(sock, SOL_SOCKET, SO_RCVBUF , &RcvBuf, sizeof(RcvBuf))==0)
    {
      RcvBuf +=1024;
    }

  DEBUGF(DebugOutput(DEBUG_OS,2, 
		     "OpenSocket> size of receive buffer now: %i (%i)\n",
                     RcvBuf,old););

  /* The used receive buffer should be smaller or equal the used buffer */
  if (RcvBuf > DG_MAXBUF) 
    {
      Printf("RcvBuf (%u) is reduced to %u\n",RcvBuf,DG_MAXBUF);
      RcvBuf = DG_MAXBUF;
    }

  DEBUGF(DebugOutput(DEBUG_OS,3,"OpenSocket> Socket created: %s!\n",
		     SocketName););

  return sock;
}


/* Taken from snort: Convert binary data to (printable) hex values */
/* this routine is only needed for debugging... */
char *
Fasthex(u_char *xdata, int length)
{
    char conv[] = "0123456789ABCDEF";
    char *retbuf = NULL; 
    char *index;
    char *end;
    char *ridx;

    index = (char *) xdata;
    end = index + length;
    retbuf = (char *) calloc((length*2)+1, sizeof(char));
    ridx = retbuf;

    while(index < end)
    {
        *ridx++ = conv[((*index & 0xFF)>>4)];
        *ridx++ = conv[((*index & 0xFF)&0x0F)];
        index++;
    }

    return retbuf;
}


/* 
 * Function: Usage
 *
 * Purpose: Display the possible command line arguments
 *
 * Arguments: argv, only needed to know the program name
 *
 * Returns: No return, simply exit...
 *
 * Author: DG
 *
 */

void 
Usage(char *argv)
{
  Printf("\nUsage: %s [-bvhl] [-A time] [-m mode] [-M mxtry] [-p port] "
	 "[-P PIDfile] [-s socket] [-S server] [-w dir] "
#ifdef DEBUGFLOP
         "[-V area,level] "
#endif
	 "[-W time ]\n",argv);
  Printf("\n -A AlarmDelay : interval in seconds to display statistics");
  Printf("\n -b            : start in daemon mode, go in background");
#ifdef DROP
  Printf("\n -D dropsocket : socket name for dropping alerts");
#endif  
  Printf("\n -h            : this information");
#ifdef DROP
  Printf("\n -H highwater  : high water mark for dropping alerts");
#endif
  Printf("\n -l            : print statistics via syslog");
#ifdef DROP
  Printf("\n -L lowwater   : low water mark for dropping alerts");
#endif
  Printf("\n -m mode       : umask value for daemon mode\n");
  Printf("\n -M maxtry     : how often should we try to connect to " 
	                     "the TCP server");
  Printf("\n -p port       : TCP port database server listens on");
  Printf("\n -P PIDfile    : file name to store pid in");
  Printf("\n -s socketname : name of unix socket to be created");
  Printf("\n -S server     : name of database server");
  Printf("\n -v            : Version information");
#ifdef DEBUGFLOP
  Printf("\n -V area,level : activates debugging, area could be 'ALL', "
	 "level up to 9");
#endif
  Printf("\n -w daemondir  : directory to change to in daemon mode");
  Printf("\n -W waittime   : time to wait between two tries to connect "
	 "to TCP server");
  Printf("\n\n");
  exit(0);
}


/* 
 * Function: CmdLine
 *
 * Purpose: Set or modify some values needed for startup
 *
 * Arguments: Commandline argv and number of elements argc
 *
 * Returns: Nothing, we set only global values
 *
 * Author: DG
 *
 */
void 
CmdLine(int argc, char* argv[]) 
{
  int opt, size, i;
#ifdef DEBUGFLOP
  uint DebugArea,DebugLevel;
  char *c;
  char options[]="bhm:vw:s:S:p:P:M:W:A:lD:L:H:V:";
#else
  char options[]="bhm:vw:s:S:p:P:M:W:A:lD:L:H:";
#endif

  while((opt = getopt(argc, argv, options)) != -1) 
    {
      switch(opt)
	{
	case 'b':
	  Daemon=1;
	  LogSyslog=1;
	  break;

	case 'm':
	  mask=(mode_t) strtol(optarg,NULL,0);
	  break;

	case 'v':
	  Printf("\nThreaded socket server to create unix socket, read\n");
	  Printf("snort data from it and transfer them via TCP socket \n");
	  Printf("to a central server.\n");
	  Printf("\nWritten by Dirk Geschke\n\n");
          Printf("%s\n\n",rcsid);
	  Printf("Version %s (snort-%s)\n\n",FLoP_VERSION,VERSION);
	  exit(0);
	  break;
      
#ifdef DROP
	case 'D':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  if (size > UNIX_PATH_MAX)
	    {
	      Printf("Length of SocketName too large!\n");
	      Printf("Maximum length: %u\n",UNIX_PATH_MAX);
	      Printf("Actual length : %u\n",size);
	      Printf("Exiting.\n");
	      exit(1);
	    }
	  strncpy(DropSocketName,optarg, size);
	  DropSocketName[size]='\0';
	  if (strlen(DropSocketName)) DropSocket=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"DropSocket: %s\n",
			     DropSocketName););
	  break;
	  
	case 'L':
	  LowWater=atoi(optarg);
	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"LowWater: %i\n",LowWater););
	  break;

	case 'H':
	  HighWater=atoi(optarg);
	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"HighWater: %i\n",HighWater););
	  break;

#endif /* DROP */

	case 'h':
	  Usage(argv[0]);
	  break;

	case 's':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  if (size > UNIX_PATH_MAX)
	    {
	      Printf("Length of SocketName too large!\n");
	      Printf("Maximum length: %u\n",UNIX_PATH_MAX);
	      Printf("Actual length : %u\n",size);
	      Printf("Exiting.\n");
	      exit(1);
	    }
	  strncpy(SocketName,optarg, size);
	  SocketName[size]='\0';

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"Socket: %s\n",SocketName););

	  break;  
	  
	case 'S':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(ServerName,optarg, size);
	  ServerName[size]='\0';

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"Server: %s\n",ServerName););

	  break;  
	  
	case 'p':
	  port=atoi(optarg);

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"Server port: %i\n",port););

	  break;  
	  
	case 'P':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(PIDFile,optarg, size);
	  PIDFile[size]='\0';

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"PID file: %s\n",PIDFile););

	  break;  

	case 'l':
	  LogSyslog=1;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "printing statistics to syslog"););

          break;

	case 'M':
	  i=atoi(optarg);
          if (i>0) MaxTry=i;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"MaxTry: %i\n",MaxTry););

	  break;  

	case 'W':
	  i=atoi(optarg);
          if (i>0) WaitTime=i;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"WaitTime: %i\n",WaitTime););

	  break;  
	  
	case 'A':
	  i=atoi(optarg);
          if (i > 0 ) AlarmDelay=i;
	  else AlarmDelay=0;

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"AlarmDelay: %i\n",WaitTime););

	  break;
	  
	case 'w':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(DaemonDir,optarg, size);
	  DaemonDir[size]='\0';

	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,"Working directory for daemon mode: %s\n",DaemonDir););

	  break;  

#ifdef DEBUGFLOP
	case 'V':
	  c=strchr(optarg,',');
	  if (c == NULL)
	    {
	      DebugArea=atoi(optarg);
	      DebugLevel=9;
	    }
	  else
	    {
	      *c='\0';c++;
	      DebugArea=atoi(optarg);
	      DebugLevel=atoi(c);
	    }

	  /* DebugLevel has maximum of 9 */
	  if (DebugLevel > 9 ) DebugLevel=9;

          /* Set debug value for area */
	  if ( (DebugArea < DEBUG_MAX) && (DebugArea > 0) )
	    {	    
	      DebugVector[DebugArea]=DebugLevel;
	      DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
				 "DebugArea: %u, DebugLevel: %u\n",
				 DebugArea,DebugLevel););
	    }
	  
	  /* if area is ascii "ALL" then set all values ( if not 
           *  already set to an value) to DebugLevel 
	   */
	  if ( DebugArea == 0 && (!strncasecmp(optarg,"ALL",3)) )
	    {
	      DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
				 "DebugArea: ALL, DebugLevel: %u\n",
				 DebugLevel););
	      for (i=0;i<DEBUG_MAX;i++)
		{
		  if (DebugVector[i] == 0 ) DebugVector[i]=DebugLevel;
		}
	    }
	  break;
#endif
	  
	default:
	  Usage(argv[0]);
	}
    }
}

/* 
 * Function: ConnectServer
 *
 * Purpose: opens TCP socket and connects to server
 *  
 * Arguments: name of (remote) server and port 
 *
 * Returns: filedescriptor of opened socket
 *
 * Author: DG
 */

int 
ConnectServer(int port)
{
#ifdef HAVE_BIND_HEADER
  int errnum;
#endif
  int  h_addr_index;
  int sock;
  int i;
  struct hostent *server_ptr;
  struct sockaddr_in srvin;
  char errbuf[STRLEN];
    
  bzero((void *)&srvin, sizeof(srvin));
  
  srvin.sin_port = htons((u_short) port);
  srvin.sin_family= AF_INET;

  DEBUGF(DebugOutput(DEBUG_CS,1,"CS> Server port: %i (%i)\n",port, 
		     srvin.sin_port););

  /* get a socket */
  sock=socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0 ) 
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("CS> socket problem: %s\n",errbuf);
      exit(1);
    }

  DEBUGF(DebugOutput(DEBUG_CS,5,"CS> Socket connected\n"););

  /* Set keepalive option for an existing connection not to    */
  /* close if actually no data is received. This is not really */
  /* necessary, otherwise we would restart a new connection    */

  DEBUGF(DebugOutput(DEBUG_CS,5,"CS> Setting keepalive option\n"););

  i=1;
  if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (void *) &i,sizeof(i)) ==-1)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("setting SO_KEEPALIVE failed: %s\n",errbuf);
    }

  /* Set reuseaddr option for a quick reuse of the used addresses */

  DEBUGF(DebugOutput(DEBUG_CS,5,"CS> Setting reuse address option\n"););

  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *) &i,sizeof(i)) ==-1)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
    Printf("setting SO_REUSEADDR failed: %s\n",errbuf);
    }
  
  /* get address of server */
#ifdef HAVE_BIND_HEADER
  server_ptr=getipnodebyname(ServerName, AF_INET, AI_ALL, &errnum);
#else
  server_ptr=gethostbyname(ServerName);
#endif
  if (server_ptr == NULL) 
    {
      Printf("%s: unknown host\n",ServerName);
      exit(1);
    }

  DEBUGF(DebugOutput(DEBUG_CS,1,"CS> Hostname: %s\n",server_ptr->h_name););

  /* open connection to server, try all possible addresses */
  /* until first call to connect() gets a connection       */
  h_addr_index=0;
  while (server_ptr->h_addr_list[h_addr_index] != NULL)
    {
      memcpy((void *)&srvin.sin_addr, server_ptr->h_addr_list[h_addr_index], 
	     server_ptr->h_length);
      if (connect(sock,(struct sockaddr*) &srvin, sizeof(srvin)) < 0) 
	{
	  if (server_ptr->h_addr_list[++h_addr_index] != NULL) 
	    {
	      /* try next address, the last one failed and */
	      /* we still have addresses...                */
	      continue;
	    }
	  /* All possible addresses failed to connect */
	  DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("failed to connect to server: %s\n",errbuf);
#ifdef HAVE_BIND_HEADER
	  freehostent(server_ptr);
#endif
          /* return a negative value to inform the calling   */
	  /* function about the failed connection. This way  */
          /* we can sleep and try it again at a later moment */ 
	  return(-1);
	}

      /* we found an adress which got a connect -> Leave loop */
      break; 
    }

#ifdef HAVE_BIND_HEADER
  freehostent(server_ptr); 
#endif

  DEBUGF(DebugOutput(DEBUG_CS,2,"CS> Connected to TCP server\n"););

  return sock;
}


/* 
 * Function: Connect
 *
 * Purpose: Wrapper for ConnectServer to try MaxTry times 
 *          to open a connection. This is for the case the
 *          server closes the connection or is missing
 *  
 * Arguments: ServerName, name of (remote) server and port 
 *            the values of number of trys to reconnect (MaxTry)
 *            and the time to wait for the next try (WaitTime) 
 *            should be free configurable (command line!)
 *            (will be done later...)
 *
 * Returns: filedescriptor of opened socket
 *
 * Author: DG
 */

int 
Connect(char *ServerName, int port)
{
  int i,ret;
  int sock=0;
  char answer[STRLEN];
  
  DEBUGF(DebugOutput(DEBUG_CONNECT,9,"Connect> Start of Connect, MaxTry: %i\n",
		     MaxTry););

  for (i=0; i<MaxTry; i++) 
    {

      DEBUGF(DebugOutput(DEBUG_CONNECT,5,
			 "Connect> Connecting to server, port: %u\n",port););

      sock=ConnectServer(port);
      if (sock <0 ) 
	{

	  DEBUGF(DebugOutput(DEBUG_CONNECT,1,
			     "Connect> Connection to server failed!\n"););
	  DEBUGF(DebugOutput(DEBUG_CONNECT,1,
			     "Connect> Sleeping for retry: %u seconds\n",
			     WaitTime););

	  sleep(WaitTime);
	}
      else 
	{
	  ReadLn(sock,answer,STRLEN);

	  DEBUGF(DebugOutput(DEBUG_CONNECT,2,"Connect> Server answer: <%s>\n",
			     answer););

	  ret=atoi(answer);
	  if (ret == 200) 
	    {
	      /* Answer from server is OK, send reply with
                 endianess of our machine */

	      if (CheckEndian()==DG_BIG_ENDIAN)
		{
		  snprintf(answer,STRLEN,"201 BigEndian");

		  DEBUGF(DebugOutput(DEBUG_CONNECT,2,
				     "Connect> Answer to send: <%s>\n",
				     answer););

		}
	      else
		{
		  snprintf(answer,STRLEN,"200 LittleEndian");

		  DEBUGF(DebugOutput(DEBUG_CONNECT,2,
				     "Connect> Answer to send: <%s>\n",
				     answer););

		}

	      if (SendLn(sock,answer,200))
		{
		  Printf("Endian mismatch -> Exiting\n");
		  exit(1);
		}
	      else
		{

		  DEBUGF(DebugOutput(DEBUG_CONNECT,2,
				     "Connect> connected!\n"););

                  set_nonblock(sock);
		  return(sock); /* we got a connection */
		}
	    }
	  else if ( ret < 400 )
	    {

	      DEBUGF(DebugOutput(DEBUG_CONNECT,2,
				 "Connect> Temporary failure, try again "
				 "in %u seconds\n",(i+1)*WaitTime););

	      /* Temporary failure, try later */
	      Printf("Server still busy, sleeping %u seconds\n",(i+1)*WaitTime);
	      sleep((i+1)*WaitTime);
	    }
	  else 
	    {
	      Printf("Serious Error with server: <%s>\n",answer);
	      exit(1);
	    }
	}
    }
  /* we were not able to get a connection to the server */
  /* all we can do is: give up and exit                 */
  if (sock <0 ) 
    {
      Printf("ERROR conneting to server -> Calling SigInt()\n");
/* 
 * We can't connect so we have a major problem. Solution: we
 * do a clean exit, we drop all buffered alerts if necessary
 * (and DROP is activated). Therefore we send ourselve a SIGINT.
 *
 */
      kill(pid,SIGINT);
      exit(1);
    }

  /* Never reached... */
  return sock;  
}


/*
 * Function: SigUsr()
 * 
 * Purpose: Print the statistics to stdout
 *
 * Author: DG
 *
 */

void 
SigUsr()
{
  static u_int RFS_last=0;
  static u_int STS_last=0;
  static u_int Drop_last=0;

#ifdef ONE_STAT_THREAD
  /* only one thread should display statistics */
  if ( !pthread_equal(RFSthread,pthread_self()) ) return;
#endif

  Printf("Entries received: %i (%i)\n",RFScounter, RFScounter-RFS_last);
  Printf("Entries send    : %i (%i)\n",STScounter, STScounter-STS_last);
  Printf("Entries dropped : %i (%i)\n",DropCounter,DropCounter-Drop_last);

  RFS_last=RFScounter;
  STS_last=STScounter;
  Drop_last=DropCounter;
}


/*
 * Function: SigInt()
 * 
 * Purpose: Catch INT-Signal to print the statistics to stdout
 *          and do a clean exit
 *
 * Author: DG
 *
 */

void 
SigInt()
{
  if ( pthread_equal(MainThread,pthread_self()) )
  {
    SigUsr();
  }
  unlink(SocketName);
  unlink(PIDFile);

#ifdef DROP
  if ( pthread_equal(RFSthread,pthread_self()) ) 
    {
      LowWater=0;
      DropPackets(data);
    }
#endif
  /* Printf("Exiting from SigInt().\n"); */
  _exit(0);
}


/*
 * Function: SigAlarm()
 * 
 * Purpose: Catch ALARM-Signal to print the statistics to stdout
 *          and do a reset of the alarm handler to get statistics
 *          on a periodically interval
 *
 * Author: DG
 *
 */

void 
SigAlarm()
{
  SigUsr();
  alarm(AlarmDelay);
}


/*
 * Function: drop2socket(data, size)
 * 
 * Purpose: sends data of size bytes to (unix domain) socket
 *
 * Returns: -1 on error
 *
 * Author: DG
 *
 */

#ifdef DROP
int 
drop2socket(void * data, int size)
{
  char errbuf[STRLEN];
  if(sendto(DropSocket,(const void *) data, size,
	    0,(struct sockaddr *)&drop_addr,sizeof(drop_addr))==-1)
    /*  if(send(UnixSocket,(const void *) data, size, 0)==-1)*/
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("error writing to drop socket <%s>: %s\n",DropSocketName,errbuf);
      return -1;
    }
  return 0;
}
#endif

#ifdef DROP
void 
DropAlert(DGPacket *data)
{
  AlertData alert;

  /* first we remove all old data and then refill all values */
  bzero((void *)&alert,sizeof(AlertData));

  snprintf((char *) alert.alertmsg, ALERTMSG_LENGTH, 
           (char *) data->alertpkt.alertmsg);
  
  alert.sid=0; /* not in database! */
  alert.cid=0;
  alert.sensorIP=SensorIP;
  alert.generator=data->alertpkt.event.sig_generator;
  alert.sig_id=data->alertpkt.event.sig_id;
  alert.reference=data->alertpkt.event.event_reference;

  alert.srcip=data->alertpkt.iph.ip_src.s_addr;
  alert.dstip=data->alertpkt.iph.ip_dst.s_addr;

  alert.protocol=data->alertpkt.iph.ip_proto;
  if (alert.protocol==IPPROTO_TCP)
    {
      alert.sport=data->alertpkt.tcph.th_sport;
      alert.dport=data->alertpkt.tcph.th_dport;
      alert.Flag.TCP.x2=data->alertpkt.tcph.th_offx2;
      alert.Flag.TCP.flags=data->alertpkt.tcph.th_flags;
    }
  else if (alert.protocol==IPPROTO_UDP)
    {
      alert.sport=data->alertpkt.udph.uh_sport;
      alert.dport=data->alertpkt.udph.uh_dport;
    }
  else if (alert.protocol==IPPROTO_ICMP)
    {
      alert.Flag.ICMP.ICMPType=data->alertpkt.icmp_code;
      alert.Flag.ICMP.ICMPCode=data->alertpkt.icmp_type;
    }
  
  alert.priority=data->alertpkt.event.priority;
  alert.classification=data->alertpkt.event.classification;
  if (data->alertpkt.val & NOPACKET_STRUCT)
    alert.time=0;
  else
    alert.time=data->alertpkt.pkth.ts.tv_sec;
  /* Finally send it to the socket */
  drop2socket((void *)&alert,sizeof(AlertData));
}
#endif /* DROP */

#ifdef DROP
void 
OpenDropSocket(char *filename)
{
  char errbuf[STRLEN];

  DEBUGF(DebugOutput(DEBUG_ODS,1,
		     "OpenDropSocket> SocketName: %s\n",filename););

  if (filename==NULL) 
    {
      Printf("Filename missing in OpenAlertSocket!\n");
      exit(1);
    }

  if(access(filename, W_OK))
    {
      Printf("WARNING: %s file doesn't exist or isn't writable!\n", filename);
    }

  bzero((char *) &drop_addr, sizeof(drop_addr));
  bcopy((const void *)filename, (void *)drop_addr.sun_path, strlen(filename));

  drop_addr.sun_family = AF_UNIX;
  
  if((DropSocket = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("ODS> socket failed: %s\n",errbuf);
      exit(1);
    }

  DEBUGF(DebugOutput(DEBUG_ODS,1,
		     "OpenDropSocket> UnixSocket: %i\n",DropSocket););
}
#endif /* DROP */

#ifdef DROP
void 
DropPackets(DGPacket *last)
{
  DGPacket *Packet;
  
  DEBUGF(DebugOutput(DEBUG_DROP,1,
		     "DropPackets> Begin to drop packets: %u\n",DropCounter););

  while ((RFScounter-DropCounter-STScounter) > LowWater)
    {
      Packet=last->prev;
      DropAlert(Packet);
      last->prev=Packet->prev;
      if (Packet->pkt_ptr) free(Packet->pkt_ptr);
      free(Packet);
      DropCounter++;
    }
  Packet=last->prev;
  if (Packet) Packet->next=last;

  DEBUGF(DebugOutput(DEBUG_DROP,1,"DropPackets> End of dropping: %u\n",
		     DropCounter););

}
#endif


/*
 *  Function: SigHup()
 *  
 *  Purpose: Catch HUP-Signal and do a restart
 * 
 *  Author: DG
 *
 */
void 
SigHup()
{
  sigset_t sigset,osigset;

  /* Acutally DROP not needed here, we allow SIGHUP's only in */
  /* the parent process, the forked processes will still work? */
#ifdef DROP
  if ( (DropSocket) && pthread_equal(RFSthread,pthread_self()) ) 
    {
      LowWater=0;
      DropPackets(data);
    }
#endif
  /* Disable alarm function */
  alarm(0);

  if ( pthread_equal(MainThread,pthread_self()) )
    {
      SigUsr();
      unlink(PIDFile);
      close(ServerSocket);
      close(UnixSocket);
      unlink(SocketName);
      Printf("Restarting %s\n",prog);

      /* Reenable SIGHUP, we were called by signal handler for SIGHUP */
      /* so normally this signal is blocked during this routine and   */
      /* this behaviour is inherited by the execv-process             */
      sigaddset(&sigset,SIGHUP);
      sigprocmask(SIG_UNBLOCK,&sigset,&osigset);
      
      /* So finally restart the main program */
      execv(prog,args);
    }
  /* never reached except if execv fails */
  _exit(1); 
}
